"""

 channel_islands_to_cct.py

 Convert the Channel Islands data set to a COCO-camera-traps .json file

 Uses the command-line tool ExifTool (exiftool.org) to pull EXIF tags from images,
 because every Python package we tried failed to pull the "Maker Notes" field properly.

"""

#%% Imports, constants, paths

## Imports ##

import os
import json
import uuid
import datetime
import glob
import subprocess
import requests
import shutil

from multiprocessing.pool import ThreadPool
from collections import defaultdict 
from urllib.parse import urlparse
from tqdm import tqdm
from PIL import Image


## Constants ##

required_input_annotation_fields = set([
    'task_id','batch_id','name','url','Object','Output','teams','task_url','Step A Agent'
    ])

n_download_threads = 10
n_exif_threads = 20
n_copy_threads = n_exif_threads


## Paths ##

input_base = r'e:\channel-islands-in'
output_base = r'g:\channel-islands-out'
exiftool_command_name = r'c:\exiftool-12.13\exiftool(-k).exe'

input_annotation_folder = os.path.join(input_base,'SCI Cameratrap Samasource Labels')
input_image_folder = os.path.join(input_base,'images')

assert os.path.isdir(input_base)
assert os.path.isdir(input_annotation_folder)
assert os.path.isdir(input_image_folder)
assert not input_annotation_folder.endswith('/')

output_file = os.path.join(output_base,'channel_islands_camera_traps.json')
output_image_folder = os.path.join(output_base,'images')
output_image_folder_humans = os.path.join(output_base,'human_images')

os.makedirs(output_base,exist_ok=True)
os.makedirs(output_image_folder,exist_ok=True)
os.makedirs(output_image_folder_humans,exist_ok=True)

# Confirm that exiftool is available
# assert which(exiftool_command_name) is not None, 'Could not locate the ExifTool executable'
assert os.path.isfile(exiftool_command_name), 'Could not locate the ExifTool executable'

parsed_input_file = os.path.join(output_base,'parsed_input.json')
exif_load_results_file = os.path.join(output_base,'exif_load_results.json')
sequence_info_results_file = os.path.join(output_base,'sequence_info_results.json')


#%% Load information from every .json file

json_files = glob.glob(input_annotation_folder+'/**/*.json', recursive=True)
print('Found {} .json files'.format(len(json_files)))

# Ignore the sample file... actually, first make sure there is a sample file
sample_files = [fn for fn in json_files if 'sample' in fn]
assert len(sample_files) == 1

# ...and now ignore that sample file.
json_files = [fn for fn in json_files if 'sample' not in fn]
input_images = []

json_basenames = set()

# json_file = json_files[0]
for json_file in tqdm(json_files):
    
    json_filename = os.path.basename(json_file)
    assert json_filename not in json_basenames
    json_basenames.add(json_filename)
    
    with open(json_file,'r') as f:        
        annotations = json.load(f)
                
    # ann = annotations[0]
    for ann in annotations:
        
        assert isinstance(ann,dict)
        ann_keys = set(ann.keys())
        assert required_input_annotation_fields == ann_keys
        ann['json_filename'] = json_filename
        input_images.append(ann)
        
    # ...for each annotation in this file

# ...for each .json file
        
print('\nLoaded {} image records from {} .json files'.format(len(input_images),len(json_files)))

image_urls = [ann['url'] for ann in input_images]


#%% Confirm URL uniqueness, handle redundant tags

output_images = []

urls = set()
for im in tqdm(input_images):
    
    url = im['url']
    
    if url in urls:        
        
        for existing_im in input_images:
            
            # Have we already added this image?
            if url == existing_im['url']:
                
                # One .json file was basically duplicated, but as:
                #
                # Ellie_2016-2017 SC12.json
                # Ellie_2016-2017-SC12.json                
                assert im['json_filename'].replace('-',' ') == existing_im['json_filename'].replace('-',' ')
                
                # If the new image has no output, just leave the old one there
                if im['Output'] is None:
                    print('Warning: duplicate URL {}, keeping existing output'.format(url))                    
                    break
                
                # If the old image has no output, and the new one has output, default to the one with output
                if (existing_im['Output'] is None) and (im['Output'] is not None): 
                    print('Warning: duplicate URL {}, adding new output'.format(url))
                    existing_im['Output'] = im['Output']
                    break
                
                else:
                    # Don't worry about the cases where someone tagged 'fox' and someone tagged 'fox_partial'
                    obj1 = im['Output'][0]['tags']['Object'].replace('_partial','')
                    obj2 = existing_im['Output'][0]['tags']['Object'].replace('_partial','')
                    if obj1 != obj2:
                        print('Warning: image {} tagged with {} and {}'.format(url,obj1,obj2))
                        
        # ...for each image we've already added
                        
    else:
        
        urls.add(url)
        output_images.append(im)
        
    # ...if this URL is/isn't in the list of URLs we've already processed

# ...for each image
        
print('Kept {} of {} annotation records'.format(len(output_images),len(input_images)))

images = output_images


#%% Save progress

with open(parsed_input_file,'w') as f:
    json.dump(images,f,indent=1)

#%%
    
if False:
    
    #%%
    
    with open(parsed_input_file,'r') as f:
        images = json.load(f)
        assert not any(['exif_tags' in im for im in images])
        

#%% Download files (functions)

# https://www.quickprogrammingtips.com/python/how-to-download-multiple-files-concurrently-in-python.html

def download_relative_url(url,overwrite=False):
    """
    Download:
        
    https://somestuff.com/my/relative/path/image.jpg
    
    ...to:
        
    [input_image_folder]/my/relative/path/image.jpg
    """
    
    parsed_url = urlparse(url)
    relative_path = parsed_url.path
    
    # This is returned with a leading slash, remove it
    relative_path = relative_path[1:]
    
    target_file = os.path.join(input_image_folder,relative_path).replace('\\','/')
    
    if os.path.isfile(target_file and not overwrite):
        print('{} exists, skipping'.format(target_file))
        return url
    
    os.makedirs(os.path.dirname(target_file),exist_ok=True)  
    
    print('Downloading {} to {}'.format(url, target_file))
    
    r = requests.get(url, stream=True)
    if r.status_code == requests.codes.ok:
        with open(target_file, 'wb') as f:
            for data in r:
                f.write(data)
    else:
        print('Warning: failed to download {}'.format(url))
        
    return url
 
    
def download_relative_urls(urls,n_threads = n_download_threads):    
    """
    Download all URLs in [urls]
    """
    if n_threads == 1:    
        results = []
        for url in urls:
            results.append(download_relative_url(url))
    else:
        results = ThreadPool(n_threads).map(download_relative_url, urls)        
    return results
    

#%% Download files (execution)
    
download_relative_urls(image_urls)


#%% Read required fields from EXIF data (functions)

def process_exif(file_path):
    """
    Get relevant fields from EXIF data for an image
    """
    
    # -G means "Print group name for each tag", e.g. print:
    #
    # [File]          Bits Per Sample                 : 8
    #
    # ...instead of:
    #
    # Bits Per Sample                 : 8
    
    proc = subprocess.Popen([exiftool_command_name, '-G', file_path], stdout=subprocess.PIPE, encoding='utf8')
    exif_lines = proc.stdout.readlines()    
    exif_lines = [s.strip() for s in exif_lines]
    assert exif_lines is not None and len(exif_lines) > 0, 'Failed to read EXIF data from {}'.format(file_path)
    
    # If we don't get any EXIF information, this probably isn't an image
    assert any([s.lower().startswith('[exif]') for s in exif_lines])
    
    exif_tags = {}
    
    found_makernotes = False
    
    # line_raw = exif_lines[0]
    for line_raw in exif_lines:
        
        line = line_raw.lower()  
        
        # Split on the first occurrence of ":"
        tokens = line.split(':',1)
        
        assert(len(tokens) == 2)
        field_name = tokens[0].strip()
        field_value = tokens[1].strip()
        
        if field_name.startswith('[makernotes]'):
            
            found_makernotes = True
            
            if 'sequence' in field_name:
                # Typically:
                #
                # '[MakerNotes]    Sequence                        ', '1 of 3']
                frame_num, seq_num_frames = field_value.split('of')
                exif_tags['frame_num'] = int(frame_num.strip())
                exif_tags['seq_num_frames'] = int(seq_num_frames.strip())
            
            # Not a typo; we are using serial number as a location
            elif 'serial number' in line:
                exif_tags['location'] = field_value            

            elif ('date/time original' in line and '[file]' not in line and '[composite]' not in line):
                
                previous_dt = None
                
                if 'datetime' in exif_tags:
                    previous_dt = exif_tags['datetime']
                dt = datetime.datetime.strptime(field_value,'%Y:%m:%d %H:%M:%S')
                
                # If there are multiple timestamps, make sure they're *almost* the same
                if previous_dt is not None:
                    delta = abs((dt-previous_dt).total_seconds())
                    assert delta < 1.0
                    
                exif_tags['datetime'] = dt
                
            if False:            
                if 'time' in line:
                    assert 'datetime' not in exif_tags
                    exif_tags['datetime'] = field_value
                    
        if (('datetime original' in line) or ('create date' in line) or ('date/time created' in line) or ('date/time original' in line)) \
            and ('[file]' not in line) and ('[composite]' not in line):
            
            previous_dt = None
            
            if 'datetime' in exif_tags:
                previous_dt = exif_tags['datetime']
            dt = datetime.datetime.strptime(field_value,'%Y:%m:%d %H:%M:%S')
            
            # If there are multiple timestamps, make sure they're *almost* the same
            if previous_dt is not None:
                delta = abs((dt-previous_dt).total_seconds())
                assert delta < 1.0
                
            exif_tags['datetime'] = dt
                
        if 'image width' in line:
            exif_tags['width'] = int(field_value)
        
        if 'image height' in line:
            exif_tags['height'] = int(field_value)
        
        if 'temperature' in line and not 'fahrenheit' in line:
            exif_tags['temperature'] = field_value
    
    # ...for each line in the exiftool output
            
    makernotes_fields = ['frame_num', 'seq_num_frames', 'location', 'temperature']
    
    if not found_makernotes:        
        
        print('Warning: could not find maker notes in {}'.format(file_path))
        
        # This isn't directly related to the lack of maker notes, but it happens that files that are missing
        # maker notes also happen to be missing EXIF date information
        if not 'datetime' in exif_tags:
            print('Warning: could not find datetime information in {}'.format(file_path))
        
        for field_name in makernotes_fields:
            assert field_name not in exif_tags
            exif_tags[field_name] = 'unknown'
            
    else:
        
        assert 'datetime' in exif_tags, 'Could not find datetime information in {}'.format(file_path)    
        for field_name in makernotes_fields:
            assert field_name in exif_tags, 'Could not find {} in {}'.format(field_name,file_path)
    
    return exif_tags

# ...process_exif()
    

def get_image_local_path(im):
    
    url = im['url']
    parsed_url = urlparse(url)
    relative_path = parsed_url.path
    
    # This is returned with a leading slash, remove it
    relative_path = relative_path[1:]    
    
    absolute_path = os.path.join(input_image_folder,relative_path).replace('\\','/')
    return absolute_path

    
def add_exif_data(im, overwrite=False):
    
    if ('exif_tags' in im) and (overwrite==False):
        return None
    
    url = im['url']
    
    # Ignore non-image files
    if url.lower().endswith('ds_store') or ('dropbox.device' in url.lower()):
        im['exif_tags'] = None
        return
    
    try:
        input_image_path = get_image_local_path(im)
        assert os.path.isfile(input_image_path)
        exif_tags = process_exif(input_image_path)
        im['exif_tags'] = exif_tags
    except Exception as e:
        s = 'Error on {}: {}'.format(url,str(e))
        print(s)
        return s
    return None


#%% Read EXIF data (execution)

if n_exif_threads == 1:        
    # ann = images[0]
    for im in tqdm(images):
        add_exif_data(im)
else:
    pool = ThreadPool(n_exif_threads)
    exif_read_results = list(tqdm(pool.imap(add_exif_data, images), total=len(images)))


#%% Save progress

with open(exif_load_results_file,'w') as f:
    
    # Use default=str to handle datetime objects
    json.dump(images, f, indent=1, default=str)

#%% 
    
if False:
    
    #%%
    
    with open(exif_load_results_file,'r') as f:
        # Not deserializing datetimes yet, will do this if I actually need to run this
        images = json.load(f)
        
    
#%% Check for EXIF read errors

for i_result,result in enumerate(exif_read_results):
    
    if result is not None:
                
        print('\nError found on image {}: {}'.format(i_result,result))
        im = images[i_result]
        file_path = get_image_local_path(im)
        assert images[i_result] == im
        result = add_exif_data(im)
        assert result is None
        print('\nFixed!\n')
        exif_read_results[i_result] = result
        
        
#%% Remove junk
        
images_out = []
for im in images:
    
    url = im['url']
    
    # Ignore non-image files
    if ('ds_store' in url.lower()) or ('dropbox.device' in url.lower()):
        continue
    images_out.append(im)
    
images = images_out
   

#%% Fill in some None values 

# ...so we can sort by datetime later, and let None's be sorted arbitrarily

for im in images:
    if 'exif_tags' not in im:
        im['exif_tags'] = None
    if 'datetime' not in im['exif_tags']:
        im['exif_tags']['datetime'] = None
        
images = sorted(images, key = lambda im: im['url'])
                

#%% Find unique locations
    
locations = set()

for ann in tqdm(images):

    assert 'exif_tags' in ann
    location = ann['exif_tags']['location']
    assert location is not None and len(location) > 0
    locations.add(location)
        

#%% Synthesize sequence information        

print('Found {} locations'.format(len(locations)))

locations = list(locations)

sequences = set()
sequence_to_images = defaultdict(list) 
images = images
max_seconds_within_sequence = 10

# Sort images by time within each location
# i_location=0; location = locations[i_location]
for i_location,location in enumerate(locations):
    
    images_this_location = [im for im in images if im['exif_tags']['location'] == location]
    sorted_images_this_location = sorted(images_this_location, key = lambda im: im['exif_tags']['datetime'])
    
    current_sequence_id = None
    next_frame_number = 0
    previous_datetime = None
        
    # previous_datetime = sorted_images_this_location[0]['datetime']
    # im = sorted_images_this_camera[1]
    for i_image,im in enumerate(sorted_images_this_location):

        # Timestamp for this image, may be None
        dt = im['exif_tags']['datetime']
        
        # Start a new sequence if:
        #
        # * This image has no timestamp
        # * This image has a frame number of zero
        # * We have no previous image timestamp
        #
        if dt is None:
            delta = None
        elif previous_datetime is None:
            delta = None
        else:
            assert isinstance(dt,datetime.datetime)
            delta = (dt - previous_datetime).total_seconds()
        
        # Start a new sequence if necessary
        if delta is None or delta > max_seconds_within_sequence:
            next_frame_number = 0
            current_sequence_id = str(uuid.uuid1())
            sequences.add(current_sequence_id)
        assert current_sequence_id is not None
        
        im['seq_id'] = current_sequence_id
        im['synthetic_frame_number'] = next_frame_number
        next_frame_number = next_frame_number + 1
        previous_datetime = dt
        sequence_to_images[im['seq_id']].append(im)
    
    # ...for each image in this location

# ...for each location


#%% Count frames in each sequence
        
print('Created {} sequences from {} images'.format(len(sequences),len(images)))

num_frames_per_sequence = {}
for seq_id in sequences:
    # images_this_sequence = [im for im in images if im['seq_id'] == seq_id]
    images_this_sequence = sequence_to_images[seq_id]
    num_frames_per_sequence[seq_id] = len(images_this_sequence)
    for im in images_this_sequence:
        im['synthetic_seq_num_frames'] = len(images_this_sequence)

    
#%% Create output filenames for each image, store original filenames
        
images_per_folder = 1000
output_paths = set()

# i_location = 0; location = locations[i_location]
for i_location,location in enumerate(locations):
    
    images_this_location = [im for im in images if im['exif_tags']['location'] == location]
    sorted_images_this_location = sorted(images_this_location, key = lambda im: im['exif_tags']['datetime'])

    # i_image = 0; im = sorted_images_this_location[i_image]
    for i_image,im in enumerate(sorted_images_this_location):
    
        url = im['url']
        parsed_url = urlparse(url)
        relative_path = parsed_url.path
        relative_path = relative_path[1:]
        im['original_relative_path'] = relative_path
        image_id = uuid.uuid1()
        im['id'] = image_id
        folder_number = i_image // images_per_folder
        image_number = i_image % images_per_folder
        output_relative_path = 'loc-' + location + '/' + '{0:03d}'.format(folder_number) + '/' + '{0:03d}'.format(image_number) + '.jpg'
        im['output_relative_path'] = output_relative_path
        assert output_relative_path not in output_paths
        output_paths.add(output_relative_path)
        
assert len(output_paths) == len(images)


#%% Save progress

with open(sequence_info_results_file,'w') as f:
    
    # Use default=str to handle datetime objects
    json.dump(images, f, indent=1, default=str)

#%% 
    
if False:
    
    #%%
    
    with open(sequence_info_results_file,'r') as f:
        images = json.load(f)
        
        
#%% Copy images to their output files (functions)

def copy_image_to_output(im):
    
    source_path = os.path.join(input_image_folder,im['original_relative_path'])
    assert(os.path.isfile(source_path))
    dest_path = os.path.join(output_image_folder,im['output_relative_path'])
    os.makedirs(os.path.dirname(dest_path),exist_ok=True)
    shutil.copyfile(source_path,dest_path)
    print('Copying {} to {}'.format(source_path,dest_path))
    return None


#%% Copy images to output files (execution)

if n_copy_threads == 1:        
    for im in tqdm(images):
        copy_image_to_output(im)
else:
    pool = ThreadPool(n_copy_threads)
    copy_image_results = list(tqdm(pool.imap(copy_image_to_output, images), total=len(images)))


#%% Rename the main image list for consistency with other scripts
    
all_image_info = images


#%% Create CCT dictionaries

def transform_bbox(coords):
    """
    Derive CCT-formatted bounding boxes from the SamaSource coordinate system.
    
    SamaSource provides a list of four points (x,y) that should make a box.
    
    CCT coordinates are absolute, with the origin at the upper-left, as x,y,w,h.
    """
    
    # Make sure this is really a box
    assert len(coords) == 4
    assert all(len(coord) == 2 for coord in coords)
    assert coords[0][1] == coords[1][1]
    assert coords[2][1] == coords[3][1]
    assert coords[0][0] == coords[2][0]
    assert coords[1][0] == coords[3][0]
    
    # Transform to CCT format
    x = coords[0][0]
    y = coords[0][1]
    h = coords[2][1] - coords[0][1]
    w = coords[1][0] - coords[0][0]
    return [x, y, w, h]

annotations = []
image_ids_to_images = {}
category_name_to_category = {}

# Force the empty category to be ID 0
empty_category = {}
empty_category['id'] = 0
empty_category['name'] = 'empty'
category_name_to_category['empty'] = empty_category
next_category_id = 1

default_annotation = {}
default_annotation['tags'] = {}
default_annotation['tags']['Object'] = None

# i_image = 0; input_im = all_image_info[0]
for i_image,input_im in tqdm(enumerate(all_image_info),total=len(all_image_info)):

    output_im = {}
    output_im['id'] = input_im['id']
    output_im['file_name'] = input_im['output_relative_path']
    output_im['seq_id'] = input_im['seq_id']
    output_im['seq_num_frames'] = input_im['synthetic_seq_num_frames']
    output_im['frame_num'] = input_im['synthetic_frame_number']
    output_im['original_relative_path'] = input_im['original_relative_path']
    
    # This issue only impacted one image that wasn't a real image, it was just a screenshot 
    # showing "no images available for this camera"
    if 'location' not in input_im['exif_tags'] or input_im['exif_tags']['location'] == 'unknown':        
        print('Warning: no location for image {}, skipping'.format(
            input_im['url']))
        continue
    output_im['location'] = input_im['exif_tags']['location']
    
    assert output_im['id'] not in image_ids_to_images
    image_ids_to_images[output_im['id']] = output_im
    
    exif_tags = input_im['exif_tags']
    
    # Convert datetime if necessary
    dt = exif_tags['datetime']
    if dt is not None and isinstance(dt,str):
        dt = datetime.datetime.strptime(dt, '%Y-%m-%d %H:%M:%S')
    
    # Process temperature if available
    output_im['temperature'] = exif_tags['temperature'] if 'temperature' in exif_tags else None
    
    # Read width and height if necessary
    w = None
    h = None
    
    if 'width' in exif_tags:    
        w = exif_tags['width']
    if 'height' in exif_tags:    
        h = exif_tags['height']
        
    output_image_full_path = os.path.join(output_image_folder,input_im['output_relative_path'])
        
    if w is None or h is None:
        pil_image = Image.open(output_image_full_path)        
        w, h = pil_image.size
        
    output_im['width'] = w
    output_im['height'] = h

    # I don't know what this field is; confirming that it's always None
    assert input_im['Object'] is None
    
    # Process object and bbox
    input_annotations = input_im['Output']
    
    if input_annotations is None:
        input_annotations = [default_annotation]
        
    # os.startfile(output_image_full_path)
    
    for i_ann,input_annotation in enumerate(input_annotations):
        
        bbox = None
        
        assert isinstance(input_annotation,dict)
        
        if input_annotation['tags']['Object'] is None:
        
            # Zero is hard-coded as the empty category, but check to be safe
            category_id = 0
            assert category_name_to_category['empty']['id'] == category_id
        
        else:
            
            # I can't figure out the 'index' field, but I'm not losing sleep about it
            # assert input_annotation['index'] == 1+i_ann
            
            points = input_annotation['points']
            assert points is not None and len(points) == 4
            bbox = transform_bbox(points)
            assert len(input_annotation['tags']) == 1 and 'Object' in input_annotation['tags']
            
            # Some annotators (but not all) included "_partial" when animals were partially obscured
            category_name = input_annotation['tags']['Object'].replace('_partial','').lower().strip()
            
            # Annotators *mostly* used 'none', but sometimes 'empty'.  'empty' is CCT-correct.
            if category_name == 'none':
                category_name = 'empty'
                
            category_id = None
            
            # If we've seen this category before...
            if category_name in category_name_to_category:
                    
                category = category_name_to_category[category_name]
                category_id = category['id'] 
              
            # If this is a new category...
            else:
                
                category_id = next_category_id
                category = {}
                category['id'] = category_id
                category['name'] = category_name
                category_name_to_category[category_name] = category
                next_category_id += 1
        
        # ...if this is an empty/non-empty annotation
                
        # Create an annotation
        annotation = {}        
        annotation['id'] = str(uuid.uuid1())
        annotation['image_id'] = output_im['id']    
        annotation['category_id'] = category_id
        annotation['sequence_level_annotation'] = False
        if bbox is not None:
            annotation['bbox'] = bbox
            
        annotations.append(annotation)
        
    # ...for each annotation on this image
        
# ...for each image

images = list(image_ids_to_images.values())
categories = list(category_name_to_category.values())
print('Loaded {} annotations in {} categories for {} images'.format(
    len(annotations),len(categories),len(images)))


#%% Change *two* annotations on images that I discovered contains a human after running MDv4

manual_human_ids = ['a07fc88a-6dd8-4d66-b552-d21d50fa39d0','285363f9-d76d-4727-b530-a6bd401bb4c7']
human_id = [cat['id'] for cat in categories if cat['name'] == 'human'][0]
for ann in tqdm(annotations):
    if ann['image_id'] in manual_human_ids:
        old_cat_id = ann['category_id']
        print('Changing annotation for image {} from {} to {}'.format(
            ann['image_id'],old_cat_id,human_id))
        ann['category_id'] = human_id
    
    
#%% Move human images

human_image_ids = set()
human_id = [cat['id'] for cat in categories if cat['name'] == 'human'][0]

# ann = annotations[0]
for ann in tqdm(annotations):
    if ann['category_id'] == human_id:
        human_image_ids.add(ann['image_id'])

print('\nFound {} human images'.format(len(human_image_ids)))

for im in tqdm(images):
    if im['id'] not in human_image_ids:
        continue
    source_path = os.path.join(output_image_folder,im['file_name'])
    if not os.path.isfile(source_path):
        continue
    target_path = os.path.join(output_image_folder_humans,im['file_name'])
    print('Moving {} to {}'.format(source_path,target_path))    
    os.makedirs(os.path.dirname(target_path),exist_ok=True)
    shutil.move(source_path,target_path)
    
    
#%% Count images by location

locations_to_images = defaultdict(list)
for im in tqdm(images):
    locations_to_images[im['location']].append(im)

    
#%% Write output

info = {}
info['year'] = 2020
info['version'] = 1.0
info['description'] = 'Camera trap data collected from the Channel Islands, California'
info['contributor'] = 'The Nature Conservancy of California'

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_file, 'w'), indent=1)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
        len(images),len(annotations),len(categories)))


#%% Validate output

from megadetector.data_management.databases import integrity_check_json_db

fn = output_file
options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = output_image_folder
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = False
    
sortedCategories, data, error = integrity_check_json_db.integrity_check_json_db(fn,options)


#%% Preview labels

from megadetector.visualization import visualize_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 159
# viz_options.classes_to_exclude = [0]
viz_options.classes_to_include = ['other']
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
html_output_file,image_db = visualize_db.visualize_db(db_path=output_file,
                                                        output_dir=os.path.join(output_base,'preview'),
                                                        image_base_dir=output_image_folder,
                                                        options=viz_options)
os.startfile(html_output_file)



"""

run_inference_with_yolov5_val.py

Runs a folder of images through MegaDetector (or another YOLOv5/YOLOv8 model) with YOLO's
val.py, converting the output to the standard MD format.  The reasons this script exists,
as an alternative to the standard run_detector_batch.py are:
    
* This script provides access to YOLO's test-time augmentation tools.
* This script serves a reference implementation: by any reasonable definition, YOLOv5's
  val.py produces the "correct" result for any image, since it matches what was used in 
  training.
* This script works for any Ultralytics detection model, including YOLOv8 models  

YOLOv5's val.py uses each file's base name as a unique identifier, which doesn't work 
when you have typical camera trap images like:

* a/b/c/RECONYX0001.JPG
* d/e/f/RECONYX0001.JPG

...both of which would just be "RECONYX0001.JPG".  So this script jumps through a bunch of 
hoops to put a symlinks in a flat folder, run YOLOv5 on that folder, and map the results back 
to the real files.

If you are running a YOLOv5 model, this script currently requires the caller to supply the path
where a working YOLOv5 install lives, and assumes that the current conda environment is all set up for 
YOLOv5.  If you are running a YOLOv8 model, the folder doesn't matter, but it assumes that ultralytics
tools are available in the current environment.

By default, this script uses symlinks to format the input images in a way that YOLO's 
val.py likes, as per above.  This requires admin privileges on Windows... actually technically this 
only requires permissions to create symbolic links, but I've never seen a case where someone has
that permission and *doesn't* have admin privileges.  If you are running this script on
Windows and you don't have admin privileges, use --no_use_symlinks, which will make copies of images,
rather than using symlinks.

"""

#%% Imports

import os
import sys
import uuid
import glob
import tempfile
import shutil
import json
import copy

from tqdm import tqdm

from megadetector.utils import path_utils
from megadetector.utils import process_utils
from megadetector.utils import string_utils

from megadetector.utils.ct_utils import is_iterable, split_list_into_fixed_size_chunks
from megadetector.utils.path_utils import path_is_abs
from megadetector.data_management import yolo_output_to_md_output
from megadetector.detection.run_detector import try_download_known_detector
from megadetector.postprocessing.combine_api_outputs import combine_api_output_files

default_image_size_with_augmentation = int(1280 * 1.3)
default_image_size_with_no_augmentation = 1280


#%% Options class

class YoloInferenceOptions:
    """
    Parameters that control the behavior of run_inference_with_yolov5_val(), including 
    the input/output filenames.
    """
    
    def __init__(self):
        
        ## Required-ish ##
        
        #: Folder of images to process (can be None if image_filename_list contains absolute paths)
        self.input_folder = None
        
        #: If this is None, [input_folder] can't be None, we'll process all images in [input_folder].
        #:            
        #: If this is not None, and [input_folder] is not None, this should be a list of relative image 
        #: paths within [input_folder] to process, or a .txt or .json file containing a list of 
        #: relative image paths.
        #:
        #: If this is not None, and [input_folder] is None, this should be a list of absolute image
        #: paths, or a .txt or .json file containing a list of absolute image paths.            
        self.image_filename_list = None
        
        #: Model filename (ending in .pt), or a well-known model name (e.g. "MDV5A")
        self.model_filename = None
        
        #: .json output file, in MD results format
        self.output_file = None
        
        
        ## Optional ##
        
        #: Required for older YOLOv5 inference, not for newer ulytralytics/YOLOv8 inference
        self.yolo_working_folder = None
        
        #: Currently 'yolov5' and 'ultralytics' are supported, and really these are proxies for
        #: "the yolov5 repo" and "the ultralytics repo".
        self.model_type = 'yolov5' 
    
        #: Image size to use; this is a single int, which in ultralytics's terminology means
        #: "scale the long side of the image to this size, and preserve aspect ratio".
        self.image_size = default_image_size_with_augmentation
        
        #: Detections below this threshold will not be included in the output file
        self.conf_thres = '0.001'
        
        #: Batch size... has no impact on results, but may create memory issues if you set
        #: this to large values
        self.batch_size = 1
        
        #: Device string: typically '0' for GPU 0, '1' for GPU 1, etc., or 'cpu'
        self.device_string = '0'
        
        #: Should we enable test-time augmentation?
        self.augment = True
        
        #: Should we enable half-precision inference?
        self.half_precision_enabled = None
        
        #: Where should we stash the temporary symlinks (or copies) used to give unique identifiers to image 
        # files?
        #:
        #: If this is None, we'll create a folder in system temp space.
        self.symlink_folder = None
        
        #: Should we use symlinks to give unique identifiers to image files (vs. copies)?
        self.use_symlinks = True
        
        #: How should we guarantee that YOLO IDs (base filenames) are unique?  Choices are:
        #:
        #: * 'verify': assume image IDs are unique, but verify and error if they're not
        #: * 'links': create symlinks (or copies, depending on use_symlinks) to enforce uniqueness
        #: * 'auto': check whether IDs are unique, create links if necessary
        self.unique_id_strategy = 'links'
        
        #: Temporary folder to stash intermediate YOLO results.
        #:
        #: If this is None, we'll create a folder in system temp space.    
        self.yolo_results_folder = None
        
        #: Should we remove the symlink folder when we're done?
        self.remove_symlink_folder = True
        
        #: Should we remove the intermediate results folder when we're done?
        self.remove_yolo_results_folder = True
        
        #: These are deliberately offset from the standard MD categories; YOLOv5
        #: needs categories IDs to start at 0.
        #:
        #: This can also be a string that points to a YOLO dataset.yaml file.
        self.yolo_category_id_to_name = {0:'animal',1:'person',2:'vehicle'}
        
        #: What should we do if the output file already exists?
        #:
        #: Can be 'error', 'skip', or 'overwrite'.
        self.overwrite_handling = 'skip'
        
        #: If True, we'll do a dry run that lets you preview the YOLO val command, without
        #: actually running it.
        self.preview_yolo_command_only = False
        
        #: By default, if any errors occur while we're copying images or creating symlinks, it's
        #: game over.  If this is True, those errors become warnings, and we plow ahead.
        self.treat_copy_failures_as_warnings = False
        
        #: Save YOLO console output
        self.save_yolo_debug_output = False
        
        #: Whether to search for images recursively within [input_folder]
        #:
        #: Ignored if a list of files is provided.
        self.recursive = True
        
        #: Maximum number of images to run in a single chunk
        self.checkpoint_frequency = None
        
    # ...def __init__()
    
# ...YoloInferenceOptions()


#%% Support functions

def _clean_up_temporary_folders(options,
                                symlink_folder,yolo_results_folder,
                                symlink_folder_is_temp_folder,yolo_folder_is_temp_folder):
    """
    Remove temporary symlink/results folders, unless the caller requested that we leave them in place.
    """
    
    if options.remove_symlink_folder:
        shutil.rmtree(symlink_folder)
    elif symlink_folder_is_temp_folder:
        print('Warning: using temporary symlink folder {}, but not removing it'.format(
            symlink_folder))
        
    if options.remove_yolo_results_folder:
        shutil.rmtree(yolo_results_folder)
    elif yolo_folder_is_temp_folder:
        print('Warning: using temporary YOLO results folder {}, but not removing it'.format(
            yolo_results_folder))

    
#%% Main function

def run_inference_with_yolo_val(options):
    """
    Runs a folder of images through MegaDetector (or another YOLOv5/YOLOv8 model) with YOLO's
    val.py, converting the output to the standard MD format.
    
    Args: 
        options (YoloInferenceOptions): all the parameters used to control this process,
            including filenames; see YoloInferenceOptions for details            
    """
        
    ##%% Input and path handling
    
    default_options = YoloInferenceOptions()
    
    for k in options.__dict__.keys():
        if k not in default_options.__dict__:
            print('Warning: unexpected variable {} in options object'.format(k))
            
    if options.model_type == 'yolov8':
        
        print('Warning: model type "yolov8" supplied, "ultralytics" is the preferred model type string for YOLOv8 models')
        options.model_type = 'ultralytics'
        
    if (options.model_type == 'yolov5') and ('yolov8' in options.model_filename.lower()):
        print('\n\n*** Warning: model type set as "yolov5", but your model filename contains "yolov8"... did you mean to use --model_type yolov8?" ***\n\n')        
    
    if options.yolo_working_folder is None:
        assert options.model_type == 'ultralytics', \
            'A working folder is required to run YOLOv5 val.py'
    else:
        assert os.path.isdir(options.yolo_working_folder), \
            'Could not find working folder {}'.format(options.yolo_working_folder)
                
    if options.half_precision_enabled is not None:
        assert options.half_precision_enabled in (0,1), \
            'Invalid value {} for --half_precision_enabled (should be 0 or 1)'.format(
                options.half_precision_enabled)
    
    # If the model filename is a known model string (e.g. "MDv5A", download the model if necessary)
    model_filename = try_download_known_detector(options.model_filename)
    
    assert os.path.isfile(model_filename), \
        'Could not find model file {}'.format(model_filename)
    
    assert (options.input_folder is not None) or (options.image_filename_list is not None), \
        'You must specify a folder and/or a file list'
        
    if options.input_folder is not None:
        assert os.path.isdir(options.input_folder), 'Could not find input folder {}'.format(
            options.input_folder)
    
    if os.path.exists(options.output_file):
        if options.overwrite_handling == 'skip':
            print('Warning: output file {} exists, skipping'.format(options.output_file))
            return
        elif options.overwrite_handling == 'overwrite':
            print('Warning: output file {} exists, overwriting'.format(options.output_file))
        elif options.overwrite_handling == 'error':
            raise ValueError('Output file {} exists'.format(options.output_file))
        else:
            raise ValueError('Unknown output handling method {}'.format(options.overwrite_handling))
    
    os.makedirs(os.path.dirname(options.output_file),exist_ok=True)
    
    if options.input_folder is not None:
        options.input_folder = options.input_folder.replace('\\','/')
        

    ##%% Other input handling
    
    if isinstance(options.yolo_category_id_to_name,str):
        
        assert os.path.isfile(options.yolo_category_id_to_name)
        yolo_dataset_file = options.yolo_category_id_to_name
        options.yolo_category_id_to_name = \
            yolo_output_to_md_output.read_classes_from_yolo_dataset_file(yolo_dataset_file)
        print('Loaded {} category mappings from {}'.format(
            len(options.yolo_category_id_to_name),yolo_dataset_file))

    temporary_folder = None
    symlink_folder_is_temp_folder = False
    yolo_folder_is_temp_folder = False
    
    job_id = str(uuid.uuid1())
    
    def get_job_temporary_folder(tf):
        if tf is not None:
            return tf
        tempdir_base = tempfile.gettempdir()
        tf = os.path.join(tempdir_base,'md_to_yolo','md_to_yolo_' + job_id)
        os.makedirs(tf,exist_ok=True)
        return tf
        
    symlink_folder = options.symlink_folder
    yolo_results_folder = options.yolo_results_folder
    
    if symlink_folder is None:
        temporary_folder = get_job_temporary_folder(temporary_folder)
        symlink_folder = os.path.join(temporary_folder,'symlinks')
        symlink_folder_is_temp_folder = True
    
    if yolo_results_folder is None:
        temporary_folder = get_job_temporary_folder(temporary_folder)
        yolo_results_folder = os.path.join(temporary_folder,'yolo_results')
        yolo_folder_is_temp_folder = True
        
    # Attach a GUID to the symlink folder, regardless of whether we created it
    symlink_folder_inner = os.path.join(symlink_folder,job_id)
    
    os.makedirs(symlink_folder_inner,exist_ok=True)
    os.makedirs(yolo_results_folder,exist_ok=True)
    

    ##%% Enumerate images
    
    image_files_relative = None
    image_files_absolute = None
    
    if options.image_filename_list is None:
        assert options.input_folder is not None and os.path.isdir(options.input_folder), \
            'Could not find input folder {}'.format(options.input_folder)
        image_files_relative = path_utils.find_images(options.input_folder,
                                                      recursive=options.recursive,
                                                      return_relative_paths=True,
                                                      convert_slashes=True)
        image_files_absolute = [os.path.join(options.input_folder,fn) for \
                                fn in image_files_relative]
    else:
        
        if is_iterable(options.image_filename_list):
            
            image_files_relative = options.image_filename_list
            
        else:
            assert isinstance(options.image_filename_list,str), \
                'Unrecognized image filename list object type: {}'.format(options.image_filename_list)
            assert os.path.isfile(options.image_filename_list), \
                'Could not find image filename list file: {}'.format(options.image_filename_list)
            ext = os.path.splitext(options.image_filename_list).lower()
            assert ext in ('.json','.txt'), \
                'Unrecognized image filename list file extension: {}'.format(options.image_filename_list)
            if ext == '.json':
                with open(options.image_filename_list,'r') as f:
                    image_files_relative = json.load(f)
                    assert is_iterable(image_files_relative)
            else:
                assert ext == '.txt'
                with open(options.image_filename_list,'r') as f:
                    image_files_relative = f.readlines()
                    image_files_relative = [s.strip() for s in image_files_relative]
        
        # ...whether the image filename list was supplied as list vs. a filename
        
        if options.input_folder is None:
            image_files_absolute = image_files_relative
        else:
            # The list should be relative filenames
            for fn in image_files_relative:
                assert not path_is_abs(fn), \
                    'When providing a folder and a list, paths in the list should be relative'
                
            image_files_absolute = \
                [os.path.join(options.input_folder,fn) for fn in image_files_relative]
        for fn in image_files_absolute:
            assert os.path.isfile(fn), 'Could not find image file {}'.format(fn)
    
    # ...whether the caller supplied a list of filenames
    
    image_files_absolute = [fn.replace('\\','/') for fn in image_files_absolute]
    del image_files_relative
    
    
    ##%% Recurse if necessary to handle checkpoints
    
    if options.checkpoint_frequency is not None and options.checkpoint_frequency > 0:
        
        chunks = split_list_into_fixed_size_chunks(image_files_absolute,options.checkpoint_frequency)
        
        chunk_output_files = []
        
        # i_chunk = 0; chunk_files_abs = chunks[i_chunk]
        for i_chunk,chunk_files_abs in enumerate(chunks):
            
            print('Processing {} images from chunk {} of {}'.format(
                len(chunk_files_abs),i_chunk,len(chunks)))
    
            chunk_options = copy.deepcopy(options)
            
            # Run each chunk without checkpointing
            chunk_options.checkpoint_frequency = None
            
            if options.input_folder is not None:
                chunk_files_relative = \
                    [os.path.relpath(fn,options.input_folder) for fn in chunk_files_abs]
                chunk_options.image_filename_list = chunk_files_relative
            else:
                chunk_options.image_filename_list = chunk_files_abs
                
            chunk_options.image_filename_list = \
                [fn.replace('\\','/') for fn in chunk_options.image_filename_list]
                
            chunk_string = 'chunk_{}'.format(str(i_chunk).zfill(5))
            chunk_options.yolo_results_folder = yolo_results_folder + '_' + chunk_string
            chunk_options.symlink_folder = symlink_folder + '_' + chunk_string
            
            # Put the output file in the parent job's scratch folder
            chunk_output_file = os.path.join(yolo_results_folder,chunk_string + '_results_md_format.json')
            chunk_output_files.append(chunk_output_file)
            chunk_options.output_file = chunk_output_file
            
            if os.path.isfile(chunk_output_file):
                
                print('Chunk output file {} exists, checking completeness'.format(chunk_output_file))
                
                with open(chunk_output_file,'r') as f:
                    chunk_results = json.load(f)
                images_in_this_chunk_results_file = [im['file'] for im in chunk_results['images']]                    
                assert len(images_in_this_chunk_results_file) == len(chunk_options.image_filename_list), \
                    'Expected {} images in chunk results file {}, found {}, possibly this is left over from a previous job?'.format(
                        len(chunk_options.image_filename_list),chunk_output_file,
                        len(images_in_this_chunk_results_file))
                for fn in images_in_this_chunk_results_file:
                    assert fn in chunk_options.image_filename_list, \
                        'Unexpected image {} in chunk results file {}, possibly this is left over from a previous job?'.format(
                            fn,chunk_output_file)
                
                print('Chunk output file {} exists and is complete, skipping this chunk'.format(
                    chunk_output_file))
                
            # ...if the outptut file exists
            
            else:
                
                run_inference_with_yolo_val(chunk_options)
                
            # ...if we do/don't have to run this chunk
            
            assert os.path.isfile(chunk_options.output_file)
            
        # ...for each chunk
    
        # Merge
        _ = combine_api_output_files(input_files=chunk_output_files,
                                 output_file=options.output_file,
                                 require_uniqueness=True,
                                 verbose=True)
        
        # Validate
        with open(options.output_file,'r') as f:
            combined_results = json.load(f)
        assert len(combined_results['images']) == len(image_files_absolute), \
            'Expected {} images in merged output file, found {}'.format(
                len(image_files_absolute),len(combined_results['images']))
        
        # Clean up
        _clean_up_temporary_folders(options,
                                    symlink_folder,yolo_results_folder,
                                    symlink_folder_is_temp_folder,yolo_folder_is_temp_folder)
        
        return
    
    # ...if we need to make recursive calls for file chunks
    
    
    ##%% Create symlinks (or copy images) to give a unique ID to each image
    
    # Maps YOLO image IDs (base filename without extension as it will appear in YOLO .json output)
    # to the *original full path* for each image (not the symlink path).
    image_id_to_file = {}  
    
    # Maps YOLO image IDs (base filename without extension as it will appear in YOLO .json output)
    # to errors, including errors that happen before we run the model at all (e.g. file access errors).
    image_id_to_error = {}
    
    create_links = True
    
    if options.unique_id_strategy == 'links':
        
        create_links = True
        
    else:
        
        assert options.unique_id_strategy in ('auto','verify'), \
            'Unknown unique ID strategy {}'.format(options.unique_id_strategy)
            
        image_ids_are_unique = True
        
        for i_image,image_fn in tqdm(enumerate(image_files_absolute),total=len(image_files_absolute)):        
            
            image_id = os.path.splitext(os.path.basename(image_fn))[0]
            
            # Is this image ID unique?
            if image_id in image_id_to_file:
                if options.unique_id_strategy == 'verify':
                    raise ValueError('"verify" specified for image uniqueness, but ' +
                                     'image ID {} occurs more than once:\n\n{}\n\n{}'.format(
                                         image_id,image_fn,image_id_to_file[image_id]))
                else:
                    assert options.unique_id_strategy == 'auto'
                    image_ids_are_unique = False
                    image_id_to_file = {}
                    break
                
            image_id_to_file[image_id] = image_fn
        
        # ...for each image
        
        if image_ids_are_unique:
            
            print('"{}" specified for image uniqueness and images are unique, skipping links'.format(
                options.unique_id_strategy))
            assert len(image_id_to_file) == len(image_files_absolute)
            create_links = False
            
        else:
            
            assert options.unique_id_strategy == 'auto'
            create_links = True
            link_type = 'copies'
            if options.use_symlinks:
                link_type = 'links'
            print('"auto" specified for image uniqueness and images are not unique, defaulting to {}'.format(
                link_type))
            
    # ...which unique ID strategy?
    
    if create_links:
        
        if options.use_symlinks:
            print('Creating {} symlinks in {}'.format(len(image_files_absolute),symlink_folder_inner))
        else:
            print('Symlinks disabled, copying {} images to {}'.format(len(image_files_absolute),symlink_folder_inner))
            
        link_full_paths = []
        
        # i_image = 0; image_fn = image_files_absolute[i_image]
        for i_image,image_fn in tqdm(enumerate(image_files_absolute),total=len(image_files_absolute)):
            
            ext = os.path.splitext(image_fn)[1]
            
            # YOLO .json output identifies images by the base filename without the extension
            image_id = str(i_image).zfill(10)
            image_id_to_file[image_id] = image_fn
            symlink_name = image_id + ext
            symlink_full_path = os.path.join(symlink_folder_inner,symlink_name)
            link_full_paths.append(symlink_full_path)
            
            try:
                
                if options.use_symlinks:
                    path_utils.safe_create_link(image_fn,symlink_full_path)
                else:
                    shutil.copyfile(image_fn,symlink_full_path)
                    
            except Exception as e:
                
                error_string = str(e)
                image_id_to_error[image_id] = error_string
                
                # Always break if the user is trying to create symlinks on Windows without
                # permission, 100% of images will always fail in this case.
                if ('a required privilege is not held by the client' in error_string.lower()) or \
                   (not options.treat_copy_failures_as_warnings):
                       
                       print('\nError copying/creating link for input file {}: {}'.format(
                           image_fn,error_string))
                       
                       raise
                       
                else:
                    
                    print('Warning: error copying/creating link for input file {}: {}'.format(
                        image_fn,error_string))
                    continue
            
            # ...except
            
        # ...for each image
                
    # ...if we need to create links/copies

    
    ##%% Create the dataset file if necessary
    
    # This may have been passed in as a string, but at this point, we should have
    # loaded the dataset file.
    assert isinstance(options.yolo_category_id_to_name,dict)
    
    # Category IDs need to be continuous integers starting at 0
    category_ids = sorted(list(options.yolo_category_id_to_name.keys()))
    assert category_ids[0] == 0
    assert len(category_ids) == 1 + category_ids[-1]
        
    yolo_dataset_file = os.path.join(yolo_results_folder,'dataset.yaml')
    yolo_image_list_file = os.path.join(yolo_results_folder,'images.txt')
    
    
    with open(yolo_image_list_file,'w') as f:
        
        if create_links:
            image_files_to_write = link_full_paths
        else:
            image_files_to_write = image_files_absolute
            
        for fn_abs in image_files_to_write:
            # At least in YOLOv5 val (need to verify for YOLOv8 val), filenames in this 
            # text file are treated as relative to the text file itself if they start with
            # "./", otherwise they're treated as absolute paths.  Since we don't want to put this
            # text file in the image folder, we'll use absolute paths.
            # fn_relative = os.path.relpath(fn_abs,options.input_folder)
            # f.write(fn_relative + '\n')
            f.write(fn_abs + '\n')
    
    if create_links:
        inference_folder = symlink_folder_inner
    else:
        # This doesn't matter, but it has to be a valid path
        inference_folder = options.yolo_results_folder
        
    with open(yolo_dataset_file,'w') as f:
        
        f.write('path: {}\n'.format(inference_folder))        
        # These need to be valid paths, even if you're not using them, and "." is always safe
        f.write('train: .\n')
        f.write('val: .\n')
        f.write('test: {}\n'.format(yolo_image_list_file))
        f.write('\n')
        f.write('nc: {}\n'.format(len(options.yolo_category_id_to_name)))
        f.write('\n')
        f.write('names:\n')
        for category_id in category_ids:
            assert isinstance(category_id,int)
            f.write('  {}: {}\n'.format(category_id,
                                        options.yolo_category_id_to_name[category_id]))


    ##%% Prepare Python command or YOLO CLI command
    
    image_size_string = str(round(options.image_size))
    
    if options.model_type == 'yolov5':
        
        cmd = 'python val.py --task test --data "{}"'.format(yolo_dataset_file)
        cmd += ' --weights "{}"'.format(model_filename)
        cmd += ' --batch-size {} --imgsz {} --conf-thres {}'.format(
            options.batch_size,image_size_string,options.conf_thres)
        cmd += ' --device "{}" --save-json'.format(options.device_string)
        cmd += ' --project "{}" --name "{}" --exist-ok'.format(yolo_results_folder,'yolo_results')
        
        if options.augment:
            cmd += ' --augment'
                
        # --half is a store_true argument for YOLOv5's val.py
        if (options.half_precision_enabled is not None) and (options.half_precision_enabled == 1):
            cmd += ' --half'
        
        # Sometimes useful for debugging
        # cmd += ' --save_conf --save_txt'
        
    elif options.model_type == 'ultralytics':
                
        if options.augment:
            augment_string = 'augment'
        else:
            augment_string = ''
            
        cmd = 'yolo val {} model="{}" imgsz={} batch={} data="{}" project="{}" name="{}" device="{}"'.\
            format(augment_string,model_filename,image_size_string,options.batch_size,
                   yolo_dataset_file,yolo_results_folder,'yolo_results',options.device_string)
        cmd += ' save_json exist_ok'
        
        if (options.half_precision_enabled is not None):
            if options.half_precision_enabled == 1:
                cmd += ' --half=True'
            else:
                assert options.half_precision_enabled == 0
                cmd += ' --half=False'
        
        # Sometimes useful for debugging
        # cmd += ' save_conf save_txt'
            
    else:
        
        raise ValueError('Unrecognized model type {}'.format(options.model_type))
        
    # print(cmd); import clipboard; clipboard.copy(cmd)

    
    ##%% Run YOLO command
    
    if options.yolo_working_folder is not None:
        current_dir = os.getcwd()
        os.chdir(options.yolo_working_folder)

    print('Running YOLO inference command:\n{}\n'.format(cmd))
    
    if options.preview_yolo_command_only:
        
        if options.remove_symlink_folder:
            try:
                print('Removing YOLO symlink folder {}'.format(symlink_folder))
                shutil.rmtree(symlink_folder)
            except Exception:
                print('Warning: error removing symlink folder {}'.format(symlink_folder))
                pass
        if options.remove_yolo_results_folder:
            try:
                print('Removing YOLO results folder {}'.format(yolo_results_folder))
                shutil.rmtree(yolo_results_folder)
            except Exception:
                print('Warning: error removing YOLO results folder {}'.format(yolo_results_folder))
                pass
        
        # sys.exit()
        return
    
    execution_result = process_utils.execute_and_print(cmd,encoding='utf-8',verbose=True)
    assert execution_result['status'] == 0, 'Error running {}'.format(options.model_type)
    yolo_console_output = execution_result['output']
      
    if options.save_yolo_debug_output:
        
        with open(os.path.join(yolo_results_folder,'yolo_console_output.txt'),'w') as f:
            for s in yolo_console_output:
                f.write(s + '\n')
        with open(os.path.join(yolo_results_folder,'image_id_to_file.json'),'w') as f:
            json.dump(image_id_to_file,f,indent=1)
        with open(os.path.join(yolo_results_folder,'image_id_to_error.json'),'w') as f:
            json.dump(image_id_to_error,f,indent=1)
                
        
    # YOLO console output contains lots of ANSI escape codes, remove them for easier parsing
    yolo_console_output = [string_utils.remove_ansi_codes(s) for s in yolo_console_output]
    
    # Find errors that occurred during the initial corruption check; these will not be included in the
    # output.  Errors that occur during inference will be handled separately.
    yolo_read_failures = []
    
    for line in yolo_console_output:
        # Lines look like:
        #
        # For ultralytics val:
        #
        # val: WARNING ⚠️ /a/b/c/d.jpg: ignoring corrupt image/label: [Errno 13] Permission denied: '/a/b/c/d.jpg'
        # line = "val: WARNING ⚠️ /a/b/c/d.jpg: ignoring corrupt image/label: [Errno 13] Permission denied: '/a/b/c/d.jpg'"
        #
        # For yolov5 val.py:
        #
        # test: WARNING: a/b/c/d.jpg: ignoring corrupt image/label: cannot identify image file '/a/b/c/d.jpg'
        # line = "test: WARNING: a/b/c/d.jpg: ignoring corrupt image/label: cannot identify image file '/a/b/c/d.jpg'"
        if 'cannot identify image file' in line:
            tokens = line.split('cannot identify image file')
            image_name = tokens[-1].strip()
            assert image_name[0] == "'" and image_name [-1] == "'"
            image_name = image_name[1:-1]
            yolo_read_failures.append(image_name)            
        elif 'ignoring corrupt image/label' in line:
            assert 'WARNING' in line
            if '⚠️' in line:
                assert line.startswith('val'), \
                    'Unrecognized line in YOLO output: {}'.format(line)
                tokens = line.split('ignoring corrupt image/label')
                image_name = tokens[0].split('⚠️')[-1].strip()
            else:
                assert line.startswith('test'), \
                    'Unrecognized line in YOLO output: {}'.format(line)
                tokens = line.split('ignoring corrupt image/label')
                image_name = tokens[0].split('WARNING:')[-1].strip()
            assert image_name.endswith(':')
            image_name = image_name[0:-1]
            yolo_read_failures.append(image_name)
                    
    # image_file = yolo_read_failures[0]
    for image_file in yolo_read_failures:
        image_id = os.path.splitext(os.path.basename(image_file))[0]
        assert image_id in image_id_to_file, 'Unexpected image ID {}'.format(image_id)
        if image_id not in image_id_to_error:
            image_id_to_error[image_id] = 'YOLO read failure'
    
    if options.yolo_working_folder is not None:
        os.chdir(current_dir)
        
    
    ##%% Convert results to MD format
    
    json_files = glob.glob(yolo_results_folder + '/yolo_results/*.json')
    assert len(json_files) == 1    
    yolo_json_file = json_files[0]

    # Map YOLO image IDs to paths
    image_id_to_relative_path = {}
    for image_id in image_id_to_file:
        fn = image_id_to_file[image_id].replace('\\','/')
        assert path_is_abs(fn)
        if options.input_folder is not None:
            assert os.path.isdir(options.input_folder)
            assert options.input_folder in fn, 'Internal error: base folder {} not in file {}'.format(
                options.input_folder,fn)
            relative_path = os.path.relpath(fn,options.input_folder)
        else:
            # We'll use the absolute path as a relative path, and pass '/'
            # as the base path in this case.
            relative_path = fn
        image_id_to_relative_path[image_id] = relative_path
        
    # Are we working with a base folder?
    if options.input_folder is not None:
        assert os.path.isdir(options.input_folder)
        image_base = options.input_folder
    else:
        image_base = '/'
        
    yolo_output_to_md_output.yolo_json_output_to_md_output(
        yolo_json_file=yolo_json_file,
        image_folder=image_base,
        output_file=options.output_file,
        yolo_category_id_to_name=options.yolo_category_id_to_name,
        detector_name=os.path.basename(model_filename),
        image_id_to_relative_path=image_id_to_relative_path,
        image_id_to_error=image_id_to_error)


    ##%% Clean up
    
    _clean_up_temporary_folders(options,
                                symlink_folder,yolo_results_folder,
                                symlink_folder_is_temp_folder,yolo_folder_is_temp_folder)
        
# ...def run_inference_with_yolo_val()


#%% Command-line driver

import argparse
from megadetector.utils.ct_utils import args_to_object

def main():
    
    options = YoloInferenceOptions()
    
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'model_filename',type=str,
        help='model file name')
    parser.add_argument(
        'input_folder',type=str,
        help='folder on which to recursively run the model')
    parser.add_argument(
        'output_file',type=str,
        help='.json file where output will be written')
    
    parser.add_argument(
        '--image_filename_list',type=str,default=None,
        help='.json or .txt file containing a list of relative image filenames within [input_folder]')
    parser.add_argument(
        '--yolo_working_folder',type=str,default=None,
        help='folder in which to execute val.py (not necessary for YOLOv8 inference)')
    parser.add_argument(
        '--image_size', default=None, type=int,
        help='image size for model execution (default {} when augmentation is enabled, else {})'.format(
            default_image_size_with_augmentation,default_image_size_with_no_augmentation))
    parser.add_argument(
        '--conf_thres', default=options.conf_thres, type=float,
        help='confidence threshold for including detections in the output file (default {})'.format(
            options.conf_thres))
    parser.add_argument(
        '--batch_size', default=options.batch_size, type=int,
        help='inference batch size (default {})'.format(options.batch_size))
    parser.add_argument(
        '--half_precision_enabled', default=None, type=int,
        help='use half-precision-inference (1 or 0) (default is the underlying model\'s default, probably full for YOLOv8 and half for YOLOv5')
    parser.add_argument(
        '--device_string', default=options.device_string, type=str,
        help='CUDA device specifier, typically "0" or "1" for CUDA devices, "mps" for M1/M2 devices, or "cpu" (default {})'.format(
            options.device_string))
    parser.add_argument(
        '--overwrite_handling', default=options.overwrite_handling, type=str,
        help='action to take if the output file exists (skip, error, overwrite) (default {})'.format(
            options.overwrite_handling))
    parser.add_argument(
        '--yolo_dataset_file', default=None, type=str,
        help='YOLOv5 dataset.yaml file from which we should load category information ' + \
            '(otherwise defaults to MD categories)')
    parser.add_argument(
        '--model_type', default=options.model_type, type=str,
        help='model type ("yolov5" or "ultralytics" ("yolov8" behaves the same as "ultralytics")) (default {})'.format(
            options.model_type))

    parser.add_argument('--unique_id_strategy', default=options.unique_id_strategy, type=str,
        help='how should we ensure that unique filenames are passed to the YOLO val script, ' + \
             'can be "verify", "auto", or "links", see options class docs for details (default {})'.format(
                 options.unique_id_strategy))
    parser.add_argument(
        '--symlink_folder', default=None, type=str,
        help='temporary folder for symlinks (defaults to a folder in the system temp dir)')
    parser.add_argument(
        '--yolo_results_folder', default=None, type=str,
        help='temporary folder for YOLO intermediate output (defaults to a folder in the system temp dir)')
    parser.add_argument(
        '--no_use_symlinks', action='store_true',
        help='copy files instead of creating symlinks when preparing the yolo input folder')
    parser.add_argument(
        '--no_remove_symlink_folder', action='store_true',
        help='don\'t remove the temporary folder full of symlinks')
    parser.add_argument(
        '--no_remove_yolo_results_folder', action='store_true',
        help='don\'t remove the temporary folder full of YOLO intermediate files')
    parser.add_argument(
        '--save_yolo_debug_output', action='store_true',
        help='write yolo console output to a text file in the results folder, along with additional debug files')
    parser.add_argument(
        '--checkpoint_frequency', default=options.checkpoint_frequency, type=int,
        help='break the job into chunks with no more than this many images (default {})'.format(
            options.checkpoint_frequency))    
    
    parser.add_argument(
        '--nonrecursive', action='store_true',
        help='Disable recursive folder processing')
    
    parser.add_argument(
        '--preview_yolo_command_only', action='store_true',
        help='don\'t run inference, just preview the YOLO inference command (still creates symlinks)')
    
    if options.augment:
        default_augment_enabled = 1
    else:
        default_augment_enabled = 0
            
    parser.add_argument(
        '--augment_enabled', default=default_augment_enabled, type=int,
        help='enable/disable augmentation (default {})'.format(default_augment_enabled))
        
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
        
    # If the caller hasn't specified an image size, choose one based on whether augmentation
    # is enabled.
    if args.image_size is None:        
        assert args.augment_enabled in (0,1), \
            'Illegal augment_enabled value {}'.format(args.augment_enabled)
        if args.augment_enabled == 1:
            args.image_size = default_image_size_with_augmentation
        else:
            args.image_size = default_image_size_with_no_augmentation
        augment_enabled_string = 'enabled'
        if not args.augment_enabled:
            augment_enabled_string = 'disabled'
        print('Augmentation is {}, using default image size {}'.format(
            augment_enabled_string,args.image_size))
        
    args_to_object(args, options)
    
    if args.yolo_dataset_file is not None:
        options.yolo_category_id_to_name = args.yolo_dataset_file
        del options.yolo_dataset_file
        
    options.recursive = (not options.nonrecursive)
    options.remove_symlink_folder = (not options.no_remove_symlink_folder)
    options.remove_yolo_results_folder = (not options.no_remove_yolo_results_folder)
    options.use_symlinks = (not options.no_use_symlinks)
    options.augment = (options.augment_enabled > 0)        
    
    del options.nonrecursive
    del options.no_remove_symlink_folder
    del options.no_remove_yolo_results_folder
    del options.no_use_symlinks
    del options.augment_enabled
        
    print(options.__dict__)
    
    run_inference_with_yolo_val(options)    

if __name__ == '__main__':
    main()


#%% Interactive driver

if False:

    #%% Run inference on a folder
    
    input_folder = r'g:\temp\tegu-val-mini'.replace('\\','/')
    model_filename = r'g:\temp\usgs-tegus-yolov5x-231003-b8-img1280-e3002-best.pt'
    output_folder = r'g:\temp\tegu-scratch'
    yolo_working_folder = r'c:\git\yolov5-tegus'
    dataset_file = r'g:\temp\dataset.yaml'
    
    # This only impacts the output file name, it's not passed to the inference functio
    job_name = 'yolo-inference-test'
    
    model_name = os.path.splitext(os.path.basename(model_filename))[0]
    
    symlink_folder = os.path.join(output_folder,'symlinks')
    yolo_results_folder = os.path.join(output_folder,'yolo_results')
    
    output_file = os.path.join(output_folder,'{}_{}-md_format.json'.format(
        job_name,model_name))
    
    options = YoloInferenceOptions()
    
    options.yolo_working_folder = yolo_working_folder
    options.input_folder = input_folder
    options.output_file = output_file
    
    pass_image_filename_list = False    
    pass_relative_paths = True
    
    if pass_image_filename_list:
        if pass_relative_paths:
            options.image_filename_list =  [
                r"val#american_cardinal#american_cardinal#CaCa#31W.01_C83#2017-2019#C90 and C83_31W.01#(05) 18AUG17 - 05SEP17 FTC AEG#MFDC1949_000065.JPG",
                r"val#american_cardinal#american_cardinal#CaCa#31W.01_C83#2017-2019#C90 and C83_31W.01#(04) 27JUL17 - 18AUG17 FTC AEG#MFDC1902_000064.JPG"
            ]   
        else:
            options.image_filename_list =  [
                r"g:/temp/tegu-val-mini/val#american_cardinal#american_cardinal#CaCa#31W.01_C83#2017-2019#C90 and C83_31W.01#(05) 18AUG17 - 05SEP17 FTC AEG#MFDC1949_000065.JPG",
                r"g:/temp/tegu-val-mini/val#american_cardinal#american_cardinal#CaCa#31W.01_C83#2017-2019#C90 and C83_31W.01#(04) 27JUL17 - 18AUG17 FTC AEG#MFDC1902_000064.JPG"
            ]
    else:
        options.image_filename_list = None    
    
    options.yolo_category_id_to_name = dataset_file
    options.augment = False
    options.conf_thres = '0.001'
    options.batch_size = 1
    options.device_string = '0'
    options.unique_id_strategy = 'auto'
    options.overwrite_handling = 'overwrite'

    if options.augment:
        options.image_size = round(1280 * 1.3)
    else:
        options.image_size = 1280
    
    options.model_filename = model_filename
    
    options.yolo_results_folder = yolo_results_folder # os.path.join(output_folder + 'yolo_results')        
    options.symlink_folder = symlink_folder # os.path.join(output_folder,'symlinks')
    options.use_symlinks = False
    
    options.remove_symlink_folder = True
    options.remove_yolo_results_folder = True
    
    options.checkpoint_frequency = 5
    
    cmd = f'python run_inference_with_yolov5_val.py {model_filename} {input_folder} ' + \
          f'{output_file} --yolo_working_folder {yolo_working_folder} ' + \
          f' --image_size {options.image_size} --conf_thres {options.conf_thres} ' + \
          f' --batch_size {options.batch_size} ' + \
          f' --symlink_folder {options.symlink_folder} --yolo_results_folder {options.yolo_results_folder} ' + \
          f' --yolo_dataset_file {options.yolo_category_id_to_name} ' + \
          f' --unique_id_strategy {options.unique_id_strategy} --overwrite_handling {options.overwrite_handling}'
      
    if not options.remove_symlink_folder:
        cmd += ' --no_remove_symlink_folder'
    if not options.remove_yolo_results_folder: 
        cmd += ' --no_remove_yolo_results_folder'
    if options.checkpoint_frequency is not None:
        cmd += f' --checkpoint_frequency {options.checkpoint_frequency}'
    if not options.use_symlinks:
        cmd += ' --no_use_symlinks'
    if not options.augment:
        cmd += ' --augment_enabled 0'
        
    print(cmd)
    execute_in_python = False
    if execute_in_python:
        run_inference_with_yolo_val(options)
    else:
        import clipboard; clipboard.copy(cmd)




from cx_Freeze import setup, Executable 
  
setup(name = "XMP Integration" , 
      version = "3.0" , 
      description = "XMP metadata writer" , 
      executables = [Executable("xmp_integration.py")])


"""

 wi_to_json

 Prepares CCT-formatted metadata based on a Wildlife Insights data export.
 
 Mostly assumes you have the images also, for validation/QA.

"""

#%% Imports and constants

import os
import json
import pandas as pd
import shutil
import uuid
import datetime
import dateutil.parser
import sys
import subprocess
import copy

from collections import defaultdict
from tqdm import tqdm

from megadetector.visualization import visualize_db
from megadetector.data_management.databases import integrity_check_json_db

organization_name = 'organization'
input_base = os.path.expanduser('~/data/' + organization_name)
image_base = os.path.join(input_base,'deployment')
image_csv = os.path.join(input_base,'images.csv')
output_json_filename = os.path.join(input_base, organization_name + '_camera_traps.json')
preview_base = os.path.expanduser('~/data/' + organization_name + '/preview')

assert os.path.isfile(image_csv)
assert os.path.isdir(image_base)

MISSING_COMMON_NAME_TOKEN = 'MISSING'

output_encoding = 'utf-8'

# Because WI filenames are GUIDs, it's not practical to page through sequences in an
# image viewer.  So we're going to (optionally) create a copy of the data set where
# images are ordered.
create_ordered_dataset = False

ordered_image_base = os.path.join(input_base,'deployment-ordered')
ordered_json_filename = os.path.join(input_base, organization_name + '_camera_traps_ordered.json')
ordered_preview_base = os.path.expanduser('~/data/' + organization_name + '/preview-ordered')

info = {}
info['year'] = 2020
info['version'] = '1.0'
info['description'] = organization_name + ' camera traps)'
info['contributor'] = organization_name
info['date_created'] = str(datetime.date.today())

def open_file(filename):
    if sys.platform == "win32":
        os.startfile(filename)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, filename])


#%% Load ground truth

images_df = pd.read_csv(image_csv)

print('Loaded {} ground truth annotations'.format(
    len(images_df)))


#%% Take everything out of Pandas

images = images_df.to_dict('records')


#%% Synthesize common names when they're not available

for im in images:
        
    if not isinstance(im['common_name'],str):
        
        # Blank rows should always have "Blank" as the common name
        assert im['is_blank'] == 0
        assert isinstance(im['genus'],str) and isinstance(im['species'],str)
        # print('Warning: missing common name for row {} ({})'.format(i_row,row['filename']))
        im['common_name'] = im['genus'].strip() + ' ' + im['species'].strip()
    
    
#%% Convert string timestamps to Python datetimes

all_locations = set()

# im = ground_truth_dicts[0]
for im in tqdm(images):
    dt = dateutil.parser.isoparse(im['timestamp'])
    assert dt.year >= 2019 and dt.year <= 2021
    im['datetime'] = dt
    
    # The field called "location" in the WI .csv file is a URL, we want to reclaim
    # the "location" keyword for CCT output
    im['url'] = im['location']
    
    # Filenames look like, e.g., N36/100EK113/06040726.JPG    
    im['location'] = im['deployment_id']
    all_locations.add(im['location'])
    

#%% Synthesize sequence information

locations = all_locations
print('Found {} locations'.format(len(locations)))

locations = list(locations)

sequences = set()
sequence_to_images = defaultdict(list) 
max_seconds_within_sequence = 10

# Sort images by time within each location
# i_location=0; location = locations[i_location]
for i_location,location in tqdm(enumerate(locations),total=len(locations)):
    
    images_this_location = [im for im in images if im['location'] == location]
    sorted_images_this_location = sorted(images_this_location, key = lambda im: im['datetime'])
    
    current_sequence_id = None
    next_frame_number = 0
    previous_datetime = None
        
    # previous_datetime = sorted_images_this_location[0]['datetime']
    # im = sorted_images_this_camera[1]
    for i_image,im in enumerate(sorted_images_this_location):

        # Timestamp for this image, may be None
        dt = im['datetime']
        
        # Start a new sequence if:
        #
        # * This image has no timestamp
        # * This image has a frame number of zero
        # * We have no previous image timestamp
        #
        if dt is None:
            delta = None
        elif previous_datetime is None:
            delta = None
        else:
            assert isinstance(dt,datetime.datetime)
            delta = (dt - previous_datetime).total_seconds()
        
        # Start a new sequence if necessary
        if delta is None or delta > max_seconds_within_sequence:
            next_frame_number = 0
            current_sequence_id = str(uuid.uuid1())
            sequences.add(current_sequence_id)
        assert current_sequence_id is not None
        
        im['seq_id'] = current_sequence_id
        im['synthetic_frame_number'] = next_frame_number
        next_frame_number = next_frame_number + 1
        previous_datetime = dt
        sequence_to_images[im['seq_id']].append(im)
    
    # ...for each image in this location

# ...for each location


#%% Create category dict and category IDs

categories_to_counts = defaultdict(int)
category_mappings = {'blank':'empty',
                     'mammal':'unknown_mammal',
                     'bird':'unknown_bird',
                     'unknown_species':'unknown'
                     }

for c in category_mappings.values():
    assert ' ' not in c
    
# im = images[0]
for im in tqdm(images):
    
    category_name = im['common_name'].lower().replace("'",'').replace(' ','_')
    if category_name in category_mappings:
        category_name = category_mappings[category_name]
    categories_to_counts[category_name] += 1
    im['category_name'] = category_name

categories_to_counts_sorted = {k: v for k, v in sorted(categories_to_counts.items(),
                                                       key=lambda item: item[1],reverse=True)}

print('\n')
for s in categories_to_counts_sorted.keys():
    print('{}: {}'.format(s,categories_to_counts_sorted[s]))


#%% Count frames in each sequence

sequence_id_to_n_frames = defaultdict(int)

for im in tqdm(images):
    seq_id = im['seq_id']
    sequence_id_to_n_frames[seq_id] = sequence_id_to_n_frames[seq_id] + 1

for im in tqdm(images):
    seq_id = im['seq_id']
    im['seq_num_frames'] = sequence_id_to_n_frames[seq_id]


#%% Build relative paths

missing_images = []

# im = images[0]
for i_image,im in enumerate(tqdm(images)):
    # Sample URL:
    #
    # gs://project-asfasdfd/deployment/21444549/asdfasdfd-616a-4d10-a921-45ac456c568a.jpg'
    relative_path = im['url'].split('/deployment/')[1]
    assert relative_path is not None and len(relative_path) > 0
    im['relative_path'] = relative_path
    
    if not os.path.isfile(os.path.join(image_base,relative_path)):
        missing_images.append(im)
    
print('{} images are missing'.format(len(missing_images)))


#%% Double check images with multiple annotations

filename_to_images = defaultdict(list)

# im = images[0]
for im in tqdm(images):
    filename_to_images[im['relative_path']].append(im)
    
filenames_with_multiple_annotations = [fn for fn in filename_to_images.keys() if len(filename_to_images[fn]) > 1]

print('\nFound {} filenames with multiple annotations'.format(len(filenames_with_multiple_annotations)))


#%% Assemble dictionaries

images_out = []
image_id_to_image = {}
annotations = []
categories = []

category_name_to_category = {}
category_id_to_category = {}

# Force the empty category to be ID 0
empty_category = {}
empty_category['name'] = 'empty'
empty_category['id'] = 0
empty_category['count'] = 0

category_id_to_category[0] = empty_category
category_name_to_category['empty'] = empty_category
categories.append(empty_category)
next_id = 1

# input_im = images[0]
for input_im in tqdm(images):

    category_name = input_im['category_name'].lower().strip()
    
    if category_name not in category_name_to_category:        

        category_id = next_id
        next_id += 1
        category = {}
        category['id'] = category_id
        category['name'] = category_name
        category['count'] = 0
        categories.append(category)
        category_name_to_category[category_name] = category
        category_id_to_category[category_id] = category

    else:
        
        category = category_name_to_category[category_name]
        
    category_id = category['id']    
    category['count'] += 1
    
    im = {}
    im['id'] = input_im['relative_path'].replace('/','_')
    im['datetime'] = str(input_im['datetime'])
    im['file_name'] = input_im['relative_path']
    im['seq_id'] = input_im['seq_id']
    im['frame_num'] = input_im['synthetic_frame_number']
    im['seq_num_frames'] = input_im['seq_num_frames']
    im['location'] = input_im['location']
    
    if im['id'] in image_id_to_image:
        # print('Warning: image ID {} ({}) has multiple annotations'.format(im['id'],im['id'].replace('_','/')))
        pass
    else:
        image_id_to_image[im['id']] = im    
        images_out.append(im)
    
    ann = {}
    
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = im['id']
    ann['category_id'] = category_id
    ann['sequence_level_annotation'] = False
    annotations.append(ann)

# ...for each image


#%% Write output .json

data = {}
data['info'] = info
data['images'] = images_out
data['annotations'] = annotations
data['categories'] = categories

with open(output_json_filename, 'w') as f:
    json.dump(data, f, indent=1)
    
print('Finished writing json to {}'.format(output_json_filename))

    
#%% Validate .json file

options = integrity_check_json_db.IntegrityCheckOptions()
options.baseDir = image_base
options.bCheckImageSizes = False
options.bCheckImageExistence = True
options.bFindUnusedImages = True

_, _, _ = integrity_check_json_db.integrity_check_json_db(output_json_filename, options)


#%% Preview labels

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 300
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = True
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.include_filename_links = True

html_output_file, _ = visualize_db.visualize_db(db_path=output_json_filename,
                                                         output_dir=preview_base,
                                                         image_base_dir=image_base,
                                                         options=viz_options)
open_file(html_output_file)
# open_file(os.path.join(image_base,'2100703/1141a545-88d2-498b-a684-7431f7aeb324.jpg'))


#%%

if create_ordered_dataset:
    
    pass

    #%% Create ordered dataset
    
    # Because WI filenames are GUIDs, it's not practical to page through sequences in an
    # image viewer.  So we're going to create a copy of the data set where images are 
    # ordered.
    
    os.makedirs(ordered_image_base,exist_ok=True)
    
    ordered_images = {}
    
    # im = images_out[0]; im
    for im in tqdm(images_out):
        im_out = copy.deepcopy(im)
        ordered_filename = im['location'] + '_' + im['seq_id'] + '_' +\
            str(im['frame_num']) + '_' + os.path.basename(im['file_name'])
        assert ordered_filename not in ordered_images
        im_out['original_file'] = im_out['file_name']
        im_out['file_name'] = ordered_filename
        ordered_images[ordered_filename] = im_out
        
    ordered_images = list(ordered_images.values())
    
    
    #%% Create ordered .json 
    
    data_ordered = copy.copy(data)
    data_ordered['images'] = ordered_images
    
    with open(ordered_json_filename, 'w') as f:
        json.dump(data_ordered, f, indent=1)
        
    print('Finished writing json to {}'.format(ordered_json_filename))
    
    
    #%% Copy files to their new locations
        
    # im = ordered_images[0]
    for im in tqdm(ordered_images):
        output_file = os.path.join(ordered_image_base,im['file_name'])
        input_file = os.path.join(image_base,im['original_file'])
        if not os.path.isfile(input_file):
            print('Warning: file {} is missing'.format(input_file))
            continue
        shutil.copyfile(input_file,output_file)
    
    original_fn_to_ordered_fn = {}
    # im = data_ordered['images'][0]
    for im in data_ordered['images']:
        original_fn_to_ordered_fn[im['original_file']] = im['file_name']    
    
    
    #%% Preview labels in the ordered dataset
    
    viz_options = visualize_db.DbVizOptions()
    viz_options.num_to_visualize = 300
    viz_options.trim_to_images_with_bboxes = False
    viz_options.add_search_links = True
    viz_options.sort_by_filename = False
    viz_options.parallelize_rendering = True
    viz_options.include_filename_links = True
    
    html_output_file, _ = visualize_db.visualize_db(db_path=ordered_json_filename,
                                                             output_dir=ordered_preview_base,
                                                             image_base_dir=ordered_image_base,
                                                             options=viz_options)
    open_file(html_output_file)
    # open_file(os.path.join(image_base,'2100703/1141a545-88d2-498b-a684-7431f7aeb324.jpg'))
        
    
    #%% Open an ordered filename from the unordered filename    
    
    unordered_filename = '2100557/54e5c751-28b4-42e3-b6d4-e8ee290228ae.jpg'
    fn = os.path.join(ordered_image_base,original_fn_to_ordered_fn[unordered_filename])
    open_file(fn)



import os
import json
import io
import random
import requests

from PIL import Image
from multiprocessing import Pool
from datetime import datetime
from requests_toolbelt import MultipartEncoder
from requests_toolbelt.multipart import decoder


ip_address = '100.100.200.200'
port = 5050

base_url = 'http://{}:{}/v1/camera-trap/sync/'.format(ip_address, port)  


def call_api(args):
    start = datetime.now()
    
    index, url, params, data, headers = args['index'],args['url'], args['params'], args['data'], args['headers']
    print('calling api: {} starttime: {}'.format(index, start))

    response = requests.post(url, params=params, data=data, headers=headers)
    elapsed_time = datetime.now() - start
    print('\napi {} status code: {}, elapsed time in seconds {}'.format(index, response.status_code, elapsed_time.total_seconds()))
    
    get_detections(response)   
    return response

def get_detections(response):
    results = decoder.MultipartDecoder.from_response(response)
    text_results = {}
    images = {}
    for part in results.parts:
        # part is a BodyPart object with b'Content-Type', and b'Content-Disposition', the later includes 'name' and 'filename' info
        headers = {}
        for k, v in part.headers.items():
            headers[k.decode(part.encoding)] = v.decode(part.encoding)
       
        if headers.get('Content-Type', None) == 'application/json':
            text_result = json.loads(part.content.decode())

    print(text_result)


def test_load(num_requests, params, max_images=1):
    requests = []
    
    # read the images anew for each request
    index = 0
    for i in range(num_requests):
        index += 1
        files = {}
        sample_input_dir = '../../../api/synchronous/sample_input/test_images'

        image_files = os.listdir(sample_input_dir)
        random.shuffle(image_files)

        num_images = 0
        for i, image_name in enumerate(image_files):
            if not image_name.lower().endswith('.jpg'):
                continue

            if num_images >= max_images:
                break
            else:
                num_images += 1

            img_path = os.path.join(sample_input_dir, image_name)
            with open(img_path, 'rb') as f:
                content = f.read()
            files[image_name] = (image_name, content, 'image/jpeg')

        m = MultipartEncoder(fields=files)
        args = {
            'index': index,
            'url': base_url + 'detect',
            'params': params,
            'data': m,
            'headers': {'Content-Type': m.content_type}
        }
        requests.append(args)
    
    print('starting', num_requests, 'threads...')
    # images are read and in each request by the time we call the API in map()
    with Pool(num_requests) as pool:
        results = pool.map(call_api, requests)

    return results


if __name__ == "__main__":
    params = {
    'min_confidence': 0.05,
    'min_rendering_confidence': 0.2,
    'render': True
    }
    
    num_requests = 10
    max_images = 1

    start = datetime.now()
    responses = test_load(num_requests, params, max_images=max_images)
    end = datetime.now()
    total_time = end - start
    print('Total time for {} requests: {}'.format(num_requests, total_time))


"""

integrity_check_json_db.py

Does some integrity-checking and computes basic statistics on a COCO Camera Traps .json file, specifically:

* Verifies that required fields are present and have the right types
* Verifies that annotations refer to valid images
* Verifies that annotations refer to valid categories
* Verifies that image, category, and annotation IDs are unique 
* Optionally checks file existence
* Finds un-annotated images
* Finds unused categories
* Prints a list of categories sorted by count

"""

#%% Constants and environment

import argparse
import json
import os
import sys

from multiprocessing.pool import ThreadPool
from operator import itemgetter
from tqdm import tqdm

from megadetector.visualization.visualization_utils import open_image
from megadetector.utils import ct_utils


#%% Classes and environment

class IntegrityCheckOptions:
    """
    Options for integrity_check_json_db()
    """
    
    def __init__(self):
        
        #: Image path; the filenames in the .json file should be relative to this folder
        self.baseDir = ''
        
        #: Should we validate the image sizes?
        self.bCheckImageSizes = False
        
        #: Should we check that all the images in the .json file exist on disk?
        self.bCheckImageExistence = False
        
        #: Should we search [baseDir] for images that are not used in the .json file?
        self.bFindUnusedImages = False
        
        #: Should we require that all images in the .json file have a 'location' field?
        self.bRequireLocation = True
        
        #: For debugging, limit the number of images we'll process
        self.iMaxNumImages = -1
        
        #: Number of threads to use for parallelization, set to <= 1 to disable parallelization
        self.nThreads = 10
        
        #: Enable additional debug output
        self.verbose = True
        
        #: Allow integer-valued image and annotation IDs (COCO uses this, CCT files use strings)
        self.allowIntIDs = False
    
# This is used in a medium-hacky way to share modified options across threads
defaultOptions = IntegrityCheckOptions()


#%% Functions

def _check_image_existence_and_size(image,options=None):
    """
    Validate the image represented in the CCT image dict [image], which should have fields:
    
    * file_name
    * width
    * height

    Args:
        image (dict): image to validate
        options (IntegrityCheckOptions): parameters impacting validation
    
    Returns:
        bool: whether this image passes validation
    """

    if options is None:        
        options = defaultOptions
    
    assert options.bCheckImageExistence
    
    filePath = os.path.join(options.baseDir,image['file_name'])
    if not os.path.isfile(filePath):
        # print('Image path {} does not exist'.format(filePath))
        return False
    
    if options.bCheckImageSizes:
        if not ('height' in image and 'width' in image):
            print('Missing image size in {}'.format(filePath))
            return False

        # width, height = Image.open(filePath).size
        pil_im = open_image(filePath)
        width,height = pil_im.size
        if (not (width == image['width'] and height == image['height'])):
            print('Size mismatch for image {}: {} (reported {},{}, actual {},{})'.format(
                    image['id'], filePath, image['width'], image['height'], width, height))
            return False
        
    return True

  
def integrity_check_json_db(jsonFile, options=None):
    """
    Does some integrity-checking and computes basic statistics on a COCO Camera Traps .json file; see
    module header comment for a list of the validation steps.
    
    Args:
        jsonFile (str): filename to validate, or an already-loaded dict
    
    Returns:
        tuple: tuple containing:
            - sortedCategories (dict): list of categories used in [jsonFile], sorted by frequency
            - data (dict): the data loaded from [jsonFile]
            - errorInfo (dict): specific validation errors
    """
    
    if options is None:       
        options = IntegrityCheckOptions()
    
    if options.bCheckImageSizes:        
        options.bCheckImageExistence = True
     
    if options.verbose:
        print(options.__dict__)
    
    if options.baseDir is None:
        options.baseDir = ''
        
    baseDir = options.baseDir
    
    
    ##%% Read .json file if necessary, integrity-check fields
    
    if isinstance(jsonFile,dict):
        
        data = jsonFile
        
    elif isinstance(jsonFile,str):
        
        assert os.path.isfile(jsonFile), '.json file {} does not exist'.format(jsonFile)
    
        if options.verbose:
            print('Reading .json {} with base dir [{}]...'.format(
                    jsonFile,baseDir))
        
        with open(jsonFile,'r') as f:
            data = json.load(f) 
            
    else:
        
        raise ValueError('Illegal value for jsonFile')
            
    images = data['images']
    annotations = data['annotations']
    categories = data['categories']
    # info = data['info']
    assert 'info' in data, 'No info struct in database'
    
    if len(baseDir) > 0:        
        assert os.path.isdir(baseDir), 'Base directory {} does not exist'.format(baseDir)
        
        
    ##%% Build dictionaries, checking ID uniqueness and internal validity as we go
    
    imageIdToImage = {}
    annIdToAnn = {}
    catIdToCat = {}
    catNameToCat = {}
    imageLocationSet = set()
    
    if options.verbose:
        print('Checking categories...')
    
    for cat in tqdm(categories):
        
        # Confirm that required fields are present
        assert 'name' in cat
        assert 'id' in cat
        
        assert isinstance(cat['id'],int), 'Illegal category ID type: [{}]'.format(str(cat['id']))
        assert isinstance(cat['name'],str), 'Illegal category name type [{}]'.format(str(cat['name']))
        
        catId = cat['id']
        catName = cat['name']
        
        # Confirm ID uniqueness
        assert catId not in catIdToCat, 'Category ID {} is used more than once'.format(catId)
        catIdToCat[catId] = cat
        cat['_count'] = 0
        
        assert catName not in catNameToCat, 'Category name {} is used more than once'.format(catName)
        catNameToCat[catName] = cat        
        
    # ...for each category
        
    if options.verbose:
        print('\nChecking images...')
    
    if options.iMaxNumImages > 0 and len(images) > options.iMaxNumImages:
        
        if options.verbose:
            print('Trimming image list to {}'.format(options.iMaxNumImages))
        images = images[0:options.iMaxNumImages]
        
    imagePathsInJson = set()
    
    sequences = set()
    
    # image = images[0]
    for image in tqdm(images):
        
        image['_count'] = 0
        
        # Confirm that required fields are present
        assert 'file_name' in image
        assert 'id' in image

        image['file_name'] = os.path.normpath(image['file_name'])
                
        imagePathsInJson.add(image['file_name'])
        
        assert isinstance(image['file_name'],str), 'Illegal image filename type'
        
        if options.allowIntIDs:
            assert isinstance(image['id'],str) or isinstance(image['id'],int), \
                'Illegal image ID type'
        else:
            assert isinstance(image['id'],str), 'Illegal image ID type'
        
        imageId = image['id']        
        
        # Confirm ID uniqueness
        assert imageId not in imageIdToImage, 'Duplicate image ID {}'.format(imageId)
        
        imageIdToImage[imageId] = image
        
        if 'height' in image:
            assert 'width' in image, 'Image with height but no width: {}'.format(image['id'])
        
        if 'width' in image:
            assert 'height' in image, 'Image with width but no height: {}'.format(image['id'])

        if options.bRequireLocation:
            assert 'location' in image, 'No location available for: {}'.format(image['id'])
            
        if 'location' in image:
            # We previously supported ints here; this should be strings now
            # assert isinstance(image['location'], str) or isinstance(image['location'], int), \
            #  'Illegal image location type'
            assert isinstance(image['location'], str)
            imageLocationSet.add(image['location'])
    
        if 'seq_id' in image:
            sequences.add(image['seq_id'])
            
        assert not ('sequence_id' in image or 'sequence' in image), 'Illegal sequence identifier'
        
    unusedFiles = []
                
    # Are we checking for unused images?
    if (len(baseDir) > 0) and options.bFindUnusedImages:    
        
        if options.verbose:
            print('\nEnumerating images...')
        
        # Recursively enumerate images
        imagePaths = []
        for root, dirs, files in os.walk(baseDir):
            for file in files:
                if file.lower().endswith(('.jpeg', '.jpg', '.png')):
                    relDir = os.path.relpath(root, baseDir)
                    relFile = os.path.join(relDir,file)
                    relFile = os.path.normpath(relFile)
                    if len(relFile) > 2 and \
                        (relFile[0:2] == './' or relFile[0:2] == '.\\'):                     
                            relFile = relFile[2:]
                    imagePaths.append(relFile)
          
        for p in imagePaths:
            if p not in imagePathsInJson:
                # print('Image {} is unused'.format(p))
                unusedFiles.append(p)
                
    validationErrors = []
    
    # Are we checking file existence and/or image size?
    if options.bCheckImageSizes or options.bCheckImageExistence:
        
        if len(baseDir) == 0:
            print('Warning: checking image sizes without a base directory, assuming "."')
         
        if options.verbose:
            print('Checking image existence and/or image sizes...')
        
        if options.nThreads is not None and options.nThreads > 1:
            pool = ThreadPool(options.nThreads)
            # results = pool.imap_unordered(lambda x: fetch_url(x,nImages), indexedUrlList)
            defaultOptions.baseDir = options.baseDir
            defaultOptions.bCheckImageSizes = options.bCheckImageSizes
            defaultOptions.bCheckImageExistence = options.bCheckImageExistence
            results = tqdm(pool.imap(_check_image_existence_and_size, images), total=len(images))
        else:
            results = []
            for im in tqdm(images):
                results.append(_check_image_existence_and_size(im,options))
                
        for iImage,r in enumerate(results):
            if not r:            
                validationErrors.append(os.path.join(options.baseDir,images[iImage]['file_name']))
                            
    # ...for each image
    
    if options.verbose:
        print('{} validation errors (of {})'.format(len(validationErrors),len(images)))
        print('Checking annotations...')
    
    nBoxes = 0
    
    for ann in tqdm(annotations):
    
        # Confirm that required fields are present
        assert 'image_id' in ann
        assert 'id' in ann
        assert 'category_id' in ann
        
        if options.allowIntIDs:
            assert isinstance(ann['id'],str) or isinstance(ann['id'],int), \
                'Illegal annotation ID type'
            assert isinstance(ann['image_id'],str) or isinstance(ann['image_id'],int), \
                'Illegal annotation image ID type'
        else:
            assert isinstance(ann['id'],str), 'Illegal annotation ID type'
            assert isinstance(ann['image_id'],str), 'Illegal annotation image ID type'
        
        assert isinstance(ann['category_id'],int), 'Illegal annotation category ID type'
        
        if 'bbox' in ann:
            nBoxes += 1
            
        annId = ann['id']        
        
        # Confirm ID uniqueness
        assert annId not in annIdToAnn
        annIdToAnn[annId] = ann
    
        # Confirm validity
        assert ann['category_id'] in catIdToCat, \
            'Category {} not found in category list'.format(ann['category_id'])
        assert ann['image_id'] in imageIdToImage, \
          'Image ID {} referred to by annotation {}, not available'.format(
            ann['image_id'],ann['id'])
    
        imageIdToImage[ann['image_id']]['_count'] += 1
        catIdToCat[ann['category_id']]['_count'] +=1 
        
    # ...for each annotation
    
    sortedCategories = sorted(categories, key=itemgetter('_count'), reverse=True)
    
    
    ##%% Print statistics
    
    if options.verbose:
    
        # Find un-annotated images and multi-annotation images
        nUnannotated = 0
        nMultiAnnotated = 0
        
        for image in images:
            if image['_count'] == 0:
                nUnannotated += 1
            elif image['_count'] > 1:
                nMultiAnnotated += 1
                
        print('Found {} unannotated images, {} images with multiple annotations'.format(
                nUnannotated,nMultiAnnotated))
        
        if (len(baseDir) > 0) and options.bFindUnusedImages:
            print('Found {} unused image files'.format(len(unusedFiles)))
            
        nUnusedCategories = 0
        
        # Find unused categories
        for cat in categories:
            if cat['_count'] == 0:
                print('Unused category: {}'.format(cat['name']))
                nUnusedCategories += 1
        
        print('Found {} unused categories'.format(nUnusedCategories))
                
        sequenceString = 'no sequence info'
        if len(sequences) > 0:
            sequenceString = '{} sequences'.format(len(sequences))
            
        print('\nDB contains {} images, {} annotations, {} bboxes, {} categories, {}\n'.format(
                len(images),len(annotations),nBoxes,len(categories),sequenceString))
    
        if len(imageLocationSet) > 0:
            print('DB contains images from {} locations\n'.format(len(imageLocationSet)))
                
        print('Categories and annotation (not image) counts:\n')
        
        for cat in sortedCategories:
            print('{:6} {}'.format(cat['_count'],cat['name']))
        
        print('')
    
    errorInfo = {}
    errorInfo['unusedFiles'] = unusedFiles
    errorInfo['validationErrors'] = validationErrors
    
    return sortedCategories, data, errorInfo

# ...def integrity_check_json_db()
    

#%% Command-line driver
    
def main():
        
    parser = argparse.ArgumentParser()
    parser.add_argument('jsonFile')
    parser.add_argument('--bCheckImageSizes', action='store_true', 
                        help='Validate image size, requires baseDir to be specified. ' + \
                             'Implies existence checking.')
    parser.add_argument('--bCheckImageExistence', action='store_true', 
                        help='Validate image existence, requires baseDir to be specified')
    parser.add_argument('--bFindUnusedImages', action='store_true', 
                        help='Check for images in baseDir that aren\'t in the database, ' + \
                             'requires baseDir to be specified')
    parser.add_argument('--baseDir', action='store', type=str, default='', 
                        help='Base directory for images')
    parser.add_argument('--bAllowNoLocation', action='store_true',
                        help='Disable errors when no location is specified for an image')
    parser.add_argument('--iMaxNumImages', action='store', type=int, default=-1, 
                        help='Cap on total number of images to check')
    parser.add_argument('--nThreads', action='store', type=int, default=10, 
                        help='Number of threads (only relevant when verifying image ' + \
                             'sizes and/or existence)')
    
    if len(sys.argv[1:])==0:
        parser.print_help()
        parser.exit()
        
    args = parser.parse_args()
    args.bRequireLocation = (not args.bAllowNoLocation)
    options = IntegrityCheckOptions()
    ct_utils.args_to_object(args, options)
    integrity_check_json_db(args.jsonFile,options)

if __name__ == '__main__':    
    main()


#%% Interactive driver(s)

if False:
    
    #%%

    """    
    python integrity_check_json_db.py ~/data/ena24.json --baseDir ~/data/ENA24 --bAllowNoLocation
    """
    
    # Integrity-check .json files for LILA
    json_files = [os.path.expanduser('~/data/ena24.json')]
    
    options = IntegrityCheckOptions()
    options.baseDir = os.path.expanduser('~/data/ENA24')
    options.bCheckImageSizes = False
    options.bFindUnusedImages = True
    options.bRequireLocation = False
    
    # options.iMaxNumImages = 10    
    
    for json_file in json_files:
        
        sortedCategories,data,_ = integrity_check_json_db(json_file, options)



"""

 awc_to_json.py

 Convert a particular .csv file from Australian Wildlife Conservancy to CCT format.

"""

#%% Constants and environment

import pandas as pd
import uuid
import json
import time
import humanfriendly
import os
import PIL

from tqdm import tqdm

from megadetector.visualization import visualize_db
from megadetector.utils import path_utils

input_metadata_file = r"D:\wildlife_data\awc\awc_imageinfo.csv"
output_file = r"D:\wildlife_data\awc\awc_imageinfo.json"
image_base = r"D:\wildlife_data\awc"
preview_base = r"D:\wildlife_data\awc\label_preview"

filename_replacements = {'D:\\Wet Tropics':'WetTropics'}
category_mappings = {'none':'empty'}


#%% Read source data

input_metadata = pd.read_csv(input_metadata_file)

print('Read {} columns and {} rows from metadata file'.format(len(input_metadata.columns),
      len(input_metadata)))


#%% Main loop over labels

startTime = time.time()

relativePathToImage = {}

images = []
annotations = []
categoryIDToCategories = {}
missingFiles = []

duplicateImageIDs = set()

# Force the empty category to be ID 0
emptyCat = {}
emptyCat['name'] = 'empty'
emptyCat['id'] = 0
categoryIDToCategories[0] = emptyCat

# iRow = 0; row = input_metadata.iloc[iRow]
for iRow,row in tqdm(input_metadata.iterrows(),total=len(input_metadata)):
    
    # ImageID,FileName,FilePath,SpeciesID,CommonName
    imageID = str(row['ImageID'])
    fn = row['FileName']
    for k in filename_replacements:
        dirName = row['FilePath'].replace(k,filename_replacements[k])
    relativePath = os.path.join(dirName,fn)
    
    # This makes an assumption of one annotation per image, which happens to be
    # true in this data set.
    if relativePath in relativePathToImage:

        im = relativePathToImage[relativePath]
        assert im['id'] == imageID
        duplicateImageIDs.add(imageID)
            
    else:
        im = {}
        im['id'] = imageID
        im['file_name'] = relativePath
        im['seq_id'] = '-1'
        images.append(im)
        relativePathToImage[relativePath] = im
        
        fullPath = os.path.join(image_base,relativePath)
        
        if not os.path.isfile(fullPath):
            
            missingFiles.append(fullPath)
        
        else:
            # Retrieve image width and height
            pilImage = PIL.Image.open(fullPath)
            width, height = pilImage.size
            im['width'] = width
            im['height'] = height

    categoryName = row['CommonName'].lower()
    if categoryName in category_mappings:
        categoryName = category_mappings[categoryName]
        
    categoryID = row['SpeciesID']
    assert isinstance(categoryID,int)
    
    if categoryID not in categoryIDToCategories:
        category = {}
        category['name'] = categoryName
        category['id'] = categoryID
        categoryIDToCategories[categoryID] = category
    else:
        assert categoryIDToCategories[categoryID]['name'] == categoryName
    
    # Create an annotation
    ann = {}
    
    # The Internet tells me this guarantees uniqueness to a reasonable extent, even
    # beyond the sheer improbability of collisions.
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = im['id']    
    ann['category_id'] = categoryID
    
    annotations.append(ann)
    
categories = list(categoryIDToCategories.values())

elapsed = time.time() - startTime
print('Finished verifying file loop in {}, {} images, {} missing images, {} repeat labels'.format(
        humanfriendly.format_timespan(elapsed), len(images), len(missingFiles), len(duplicateImageIDs)))    


#%% Check for images that aren't included in the metadata file

# Enumerate all images
# list(relativePathToImage.keys())[0]

imageFullPaths = path_utils.find_images(image_base,bRecursive=True)
unmatchedFiles = []

for iImage,imagePath in enumerate(imageFullPaths):
    
    fn = os.path.relpath(imagePath,image_base)    
    if fn not in relativePathToImage:
        unmatchedFiles.append(fn)

print('Finished checking {} images to make sure they\'re in the metadata, found {} mismatches'.format(
        len(imageFullPaths),len(unmatchedFiles)))


#%% Create info struct

info = {}
info['year'] = 2019
info['version'] = 1
info['description'] = 'COCO style database'
info['secondary_contributor'] = 'Converted to COCO .json by Dan Morris'
info['contributor'] = ''


#%% Write output

json_data = {}
json_data['images'] = images
json_data['annotations'] = annotations
json_data['categories'] = categories
json_data['info'] = info
json.dump(json_data, open(output_file,'w'), indent=4)

print('Finished writing .json file with {} images, {} annotations, and {} categories'.format(
        len(images),len(annotations),len(categories)))


#%% Validate the database's integrity

from megadetector.data_management.databases import integrity_check_json_db

options = integrity_check_json_db.IntegrityCheckOptions()
sortedCategories,data = integrity_check_json_db.integrity_check_json_db(output_file, options)

    
#%% Render a bunch of images to make sure the labels got carried along correctly

bbox_db_path = output_file
output_dir = preview_base

options = visualize_db.BboxDbVizOptions()
options.num_to_visualize = 1000
options.sort_by_filename = False

htmlOutputFile = visualize_db.visualize_db(bbox_db_path,output_dir,image_base,options)




"""

process_video.py

Splits a video (or folder of videos) into frames, runs the frames through run_detector_batch.py,
and optionally stitches together results into a new video with detection boxes.

Operates by separating the video into frames, typically sampling every Nth frame, and writing
those frames to disk, before running MD.  This approach clearly has a downside: it requires
a bunch more disk space, compared to extracting frames and running MD on them without ever
writing them to disk.  The upside, though, is that this approach allows you to run repeat
detection elimination after running MegaDetector, and it allows allows more efficient re-use
of frames if you end up running MD more than once, or running multiple versions of MD.

"""

#%% Imports

import os
import sys
import tempfile
import argparse
import itertools
import json
import shutil
import getpass

from uuid import uuid1

from megadetector.detection import run_detector_batch
from megadetector.visualization import visualize_detector_output
from megadetector.utils.ct_utils import args_to_object
from megadetector.utils.path_utils import insert_before_extension, clean_path
from megadetector.detection.video_utils import video_to_frames
from megadetector.detection.video_utils import frames_to_video
from megadetector.detection.video_utils import frame_results_to_video_results
from megadetector.detection.video_utils import _add_frame_numbers_to_results
from megadetector.detection.video_utils import video_folder_to_frames
from megadetector.detection.video_utils import default_fourcc


#%% Classes

class ProcessVideoOptions:
    """
    Options controlling the behavior of process_video()
    """
    
    def __init__(self):
        
        #: Can be a model filename (.pt or .pb) or a model name (e.g. "MDV5A")
        #: 
        #: Use the string "no_detection" to indicate that you only want to extract frames,
        #: not run a model.  If you do this, you almost definitely want to set 
        #: keep_extracted_frames to "True", otherwise everything in this module is a no-op.
        #: I.e., there's no reason to extract frames, do nothing with them, then delete them.
        self.model_file = 'MDV5A'
        
        #: Video (of folder of videos) to process
        self.input_video_file = ''
    
        #: .json file to which we should write results
        self.output_json_file = None
        
        #: File to which we should write a video with boxes, only relevant if 
        #: render_output_video is True
        self.output_video_file = None
        
        #: Folder to use for extracted frames; will use a folder in system temp space
        #: if this is None
        self.frame_folder = None
        
        #: Folder to use for rendered frames (if rendering output video); will use a folder 
        #: in system temp space if this is None
        self.frame_rendering_folder = None
        
        #: Should we render a video with detection boxes?
        #:
        #: Only supported when processing a single video, not a folder.
        self.render_output_video = False
        
        #: If we are rendering boxes to a new video, should we keep the temporary
        #: rendered frames?
        self.keep_rendered_frames = False
        
        #: Should we keep the extracted frames?
        self.keep_extracted_frames = False
        
        #: Should we delete the entire folder the extracted frames are written to?
        #:
        #: By default, we delete the frame files but leave the (probably-empty) folder in place, 
        #: for no reason other than being paranoid about deleting folders.
        self.force_extracted_frame_folder_deletion = False
        
        #: Should we delete the entire folder the rendered frames are written to?
        #:
        #: By default, we delete the frame files but leave the (probably-empty) folder in place,
        #: for no reason other than being paranoid about deleting folders.
        self.force_rendered_frame_folder_deletion = False
         
        #: If we've already run MegaDetector on this video or folder of videos, i.e. if we 
        #: find a corresponding MD results file, should we re-use it?  Defaults to reprocessing.
        self.reuse_results_if_available = False
        
        #: If we've already split this video or folder of videos into frames, should we 
        #: we re-use those extracted frames?  Defaults to reprocessing.
        self.reuse_frames_if_available = False
        
        #: If [input_video_file] is a folder, should we search for videos recursively?
        self.recursive = False 
        
        #: Enable additional debug console output
        self.verbose = False
        
        #: fourcc code to use for writing videos; only relevant if render_output_video is True
        self.fourcc = None
    
        #: force a specific frame rate for output videos; only relevant if render_output_video 
        #: is True
        self.rendering_fs = None
        
        #: Confidence threshold to use for writing videos with boxes, only relevant if
        #: if render_output_video is True.  Defaults to choosing a reasonable threshold
        #: based on the model version.
        self.rendering_confidence_threshold = None
        
        #: Detections below this threshold will not be included in the output file.
        self.json_confidence_threshold = 0.005
        
        #: Sample every Nth frame; set to None (default) or 1 to sample every frame.  Typically
        #: we sample down to around 3 fps, so for typical 30 fps videos, frame_sample=10 is a 
        #: typical value.  Mutually exclusive with [frames_to_extract].
        self.frame_sample = None
        
        #: Extract a specific set of frames (list of ints, or a single int).  Mutually exclusive with
        #: [frame_sample].
        self.frames_to_extract = None
        
        #: Number of workers to use for parallelization; set to <= 1 to disable parallelization
        self.n_cores = 1
    
        #: For debugging only, stop processing after a certain number of frames.
        self.debug_max_frames = -1
        
        #: File containing non-standard categories, typically only used if you're running a non-MD
        #: detector.
        self.class_mapping_filename = None
        
        #: JPEG quality for frame output, from 0-100.  Defaults to the opencv default (typically 95)
        self.quality = 90
        
        #: Resize frames so they're at most this wide
        self.max_width = None
        
        #: Run the model at this image size (don't mess with this unless you know what you're
        #: getting into)
        self.image_size = None
        
        #: Enable image augmentation
        self.augment = False
    
# ...class ProcessVideoOptions


#%% Functions

def _select_temporary_output_folders(options):
    """
    Choose folders in system temp space for writing temporary frames.  Does not create folders,
    just defines them.
    """
    
    tempdir = os.path.join(tempfile.gettempdir(), 'process_camera_trap_video')
    
    # If we create a folder like "process_camera_trap_video" in the system temp dir, it may
    # be the case that no one else can write to it, even to create user-specific subfolders.
    # If we create a uuid-named folder in the system temp dir, we make a mess.  
    #
    # Compromise with "process_camera_trap_video-[user]".
    user_tempdir = tempdir + '-' + getpass.getuser()
    
    # I don't know whether it's possible for a username to contain characters that are
    # not valid filename characters, but just to be sure...
    user_tempdir = clean_path(user_tempdir)
    
    frame_output_folder = os.path.join(
        user_tempdir, os.path.basename(options.input_video_file) + '_frames_' + str(uuid1()))
    
    rendering_output_folder = os.path.join(
        tempdir, os.path.basename(options.input_video_file) + '_detections_' + str(uuid1()))
        
    temporary_folder_info = \
    {
        'temp_folder_base':user_tempdir,
        'frame_output_folder':frame_output_folder,
        'rendering_output_folder':rendering_output_folder
    }
    
    return temporary_folder_info

# ...def _create_frame_output_folders(...)


def _clean_up_rendered_frames(options,rendering_output_folder,detected_frame_files):
    """
    If necessary, delete rendered frames and/or the entire rendering output folder.
    """
    
    caller_provided_rendering_output_folder = (options.frame_rendering_folder is not None)
    
    # (Optionally) delete the temporary directory we used for rendered detection images
    if not options.keep_rendered_frames:
        
        try:
            
            # If (a) we're supposed to delete the temporary rendering folder no
            # matter where it is and (b) we created it in temp space, delete the 
            # whole tree
            if options.force_rendered_frame_folder_deletion and \
               (not caller_provided_rendering_output_folder):
                   
                if options.verbose:
                    print('Recursively deleting rendered frame folder {}'.format(
                        rendering_output_folder))
                    
                shutil.rmtree(rendering_output_folder)
                
            # ...otherwise just delete the frames, but leave the folder in place
            else:
                
                if options.force_rendered_frame_folder_deletion:
                    assert caller_provided_rendering_output_folder
                    print('Warning: force_rendered_frame_folder_deletion supplied with a ' + \
                          'user-provided folder, only removing frames')
                        
                for rendered_frame_fn in detected_frame_files:
                    os.remove(rendered_frame_fn)
                    
        except Exception as e:
            print('Warning: error deleting rendered frames from folder {}:\n{}'.format(
                rendering_output_folder,str(e)))
            pass

    elif options.force_rendered_frame_folder_deletion:
        
        print('Warning: keep_rendered_frames and force_rendered_frame_folder_deletion both ' + \
              'specified, not deleting')
            
# ...def _clean_up_rendered_frames(...)


def _clean_up_extracted_frames(options,frame_output_folder,frame_filenames):
    """
    If necessary, delete extracted frames and/or the entire temporary frame folder.
    """
    
    caller_provided_frame_output_folder = (options.frame_folder is not None)
                                           
    if not options.keep_extracted_frames:
        
        try:
            
            # If (a) we're supposed to delete the temporary frame folder no
            # matter where it is and (b) we created it in temp space, delete the 
            # whole tree.
            if options.force_extracted_frame_folder_deletion and \
               (not caller_provided_frame_output_folder):
                   
                if options.verbose:
                    print('Recursively deleting frame output folder {}'.format(frame_output_folder))
                    
                shutil.rmtree(frame_output_folder)
                
            # ...otherwise just delete the frames, but leave the folder in place
            else:
                
                if options.force_extracted_frame_folder_deletion:
                    assert caller_provided_frame_output_folder
                    print('Warning: force_extracted_frame_folder_deletion supplied with a ' + \
                          'user-provided folder, only removing frames')
                    
                for extracted_frame_fn in frame_filenames:
                    os.remove(extracted_frame_fn)
                    
        except Exception as e:
            print('Warning: error removing extracted frames from folder {}:\n{}'.format(
                frame_output_folder,str(e)))
            pass
    
    elif options.force_extracted_frame_folder_deletion:
        
        print('Warning: keep_extracted_frames and force_extracted_frame_folder_deletion both ' + \
              'specified, not deleting')

# ...def _clean_up_extracted_frames


def process_video(options):
    """
    Process a single video through MD, optionally writing a new video with boxes.
    Can also be used just to split a video into frames, without running a model.
    
    Args: 
        options (ProcessVideoOptions): all the parameters used to control this process,
            including filenames; see ProcessVideoOptions for details
            
    Returns:
        dict: frame-level MegaDetector results, identical to what's in the output .json file
    """

    if options.output_json_file is None:
        options.output_json_file = options.input_video_file + '.json'

    if options.render_output_video and (options.output_video_file is None):
        options.output_video_file = options.input_video_file + '.detections.mp4'

    if options.model_file == 'no_detection' and not options.keep_extracted_frames:
        print('Warning: you asked for no detection, but did not specify keep_extracted_frames, this is a no-op')
        return
    
    # Track whether frame and rendering folders were created by this script
    caller_provided_frame_output_folder = (options.frame_folder is not None)
    caller_provided_rendering_output_folder = (options.frame_rendering_folder is not None)
    
    # This does not create any folders, just defines temporary folder names in 
    # case we need them.
    temporary_folder_info = _select_temporary_output_folders(options)
        
    if (caller_provided_frame_output_folder):
        frame_output_folder = options.frame_folder
    else:
        frame_output_folder = temporary_folder_info['frame_output_folder']
   
    os.makedirs(frame_output_folder, exist_ok=True)


    ## Extract frames
    
    frame_filenames, Fs = video_to_frames(
                            options.input_video_file, 
                            frame_output_folder, 
                            every_n_frames=options.frame_sample, 
                            overwrite=(not options.reuse_frames_if_available),
                            quality=options.quality, 
                            max_width=options.max_width, 
                            verbose=options.verbose,
                            frames_to_extract=options.frames_to_extract)

    image_file_names = frame_filenames
    if options.debug_max_frames > 0:
        image_file_names = image_file_names[0:options.debug_max_frames]
    
    if options.model_file == 'no_detection':
        assert options.keep_extracted_frames, \
            'Internal error: keep_extracted_frames not set, but no model specified'
        return
    
    
    ## Run MegaDetector
    
    if options.reuse_results_if_available and \
        os.path.isfile(options.output_json_file):
            print('Loading results from {}'.format(options.output_json_file))
            with open(options.output_json_file,'r') as f:
                results = json.load(f)
    else:
        results = run_detector_batch.load_and_run_detector_batch(
            options.model_file, 
            image_file_names,
            confidence_threshold=options.json_confidence_threshold,
            n_cores=options.n_cores,
            class_mapping_filename=options.class_mapping_filename,
            quiet=True,
            augment=options.augment,
            image_size=options.image_size)
        
        _add_frame_numbers_to_results(results)
        
        run_detector_batch.write_results_to_file(
            results, options.output_json_file,
            relative_path_base=frame_output_folder,
            detector_file=options.model_file,
            custom_metadata={'video_frame_rate':Fs})

    
    ## (Optionally) render output video
    
    if options.render_output_video:
        
        # Render detections to images
        if (caller_provided_rendering_output_folder):
            rendering_output_dir = options.frame_rendering_folder
        else:
            rendering_output_dir = temporary_folder_info['rendering_output_folder']
            
        os.makedirs(rendering_output_dir,exist_ok=True)
        
        detected_frame_files = visualize_detector_output.visualize_detector_output(
            detector_output_path=options.output_json_file,
            out_dir=rendering_output_dir,
            images_dir=frame_output_folder,
            confidence_threshold=options.rendering_confidence_threshold)

        # Combine into a video
        if options.rendering_fs is not None:
            rendering_fs = options.rendering_fs
        elif options.frame_sample is None:
            rendering_fs = Fs
        else:
            # If the original video was 30fps and we sampled every 10th frame, 
            # render at 3fps
            rendering_fs = Fs / options.frame_sample
            
        print('Rendering {} frames to {} at {} fps (original video {} fps)'.format(
            len(detected_frame_files), options.output_video_file,rendering_fs,Fs))
        frames_to_video(detected_frame_files, 
                        rendering_fs, 
                        options.output_video_file, 
                        codec_spec=options.fourcc)
        
        # Possibly clean up rendered frames
        _clean_up_rendered_frames(options,rendering_output_dir,detected_frame_files)
    
    # ...if we're rendering video
    
    
    ## (Optionally) delete the extracted frames
    _clean_up_extracted_frames(options, frame_output_folder, frame_filenames)
    
# ...process_video()


def process_video_folder(options):
    """
    Process a folder of videos through MD. Can also be used just to split a folder of
    videos into frames, without running a model.
    
    When this function is used to run MD, two .json files will get written, one with 
    an entry for each *frame* (identical to what's created by process_video()), and
    one with an entry for each *video* (which is more suitable for, e.g., reading into
    Timelapse).
    
    Args: 
        options (ProcessVideoOptions): all the parameters used to control this process,
            including filenames; see ProcessVideoOptions for details            
    """
    
    ## Validate options

    assert os.path.isdir(options.input_video_file), \
        '{} is not a folder'.format(options.input_video_file)
           
    if options.model_file == 'no_detection' and not options.keep_extracted_frames:
        print('Warning: you asked for no detection, but did not specify keep_extracted_frames, this is a no-op')
        return
    
    if options.model_file != 'no_detection':
        assert options.output_json_file is not None, \
            'When processing a folder, you must specify an output .json file'      
        assert options.output_json_file.endswith('.json')
        video_json = options.output_json_file
        frames_json = options.output_json_file.replace('.json','.frames.json')
        os.makedirs(os.path.dirname(video_json),exist_ok=True)
    
    # Track whether frame and rendering folders were created by this script
    caller_provided_frame_output_folder = (options.frame_folder is not None)
    caller_provided_rendering_output_folder = (options.frame_rendering_folder is not None)
    
    # This does not create any folders, just defines temporary folder names in 
    # case we need them.
    temporary_folder_info = _select_temporary_output_folders(options)
    
    
    ## Split every video into frames
    
    if caller_provided_frame_output_folder:
        frame_output_folder = options.frame_folder
    else:
        frame_output_folder = temporary_folder_info['frame_output_folder']
        
    os.makedirs(frame_output_folder, exist_ok=True)

    print('Extracting frames')
    
    frame_filenames, Fs, video_filenames = \
        video_folder_to_frames(input_folder=options.input_video_file,
                               output_folder_base=frame_output_folder, 
                               recursive=options.recursive, 
                               overwrite=(not options.reuse_frames_if_available),
                               n_threads=options.n_cores,
                               every_n_frames=options.frame_sample,
                               verbose=options.verbose,
                               quality=options.quality,
                               max_width=options.max_width,
                               frames_to_extract=options.frames_to_extract)
    
    print('Extracted frames for {} videos'.format(len(set(video_filenames))))
    image_file_names = list(itertools.chain.from_iterable(frame_filenames))
    
    if len(image_file_names) == 0:
        if len(video_filenames) == 0:
            print('No videos found in folder {}'.format(options.input_video_file))
        else:
            print('No frames extracted from folder {}, this may be due to an '\
                  'unsupported video codec'.format(options.input_video_file))
        return

    if options.debug_max_frames is not None and options.debug_max_frames > 0:
        image_file_names = image_file_names[0:options.debug_max_frames]
        
    if options.model_file == 'no_detection':
        assert options.keep_extracted_frames, \
            'Internal error: keep_extracted_frames not set, but no model specified'
        return
    
    
    ## Run MegaDetector on the extracted frames
    
    if options.reuse_results_if_available and \
        os.path.isfile(frames_json):
            print('Bypassing inference, loading results from {}'.format(frames_json))
            results = None
    else:
        print('Running MegaDetector')
        results = run_detector_batch.load_and_run_detector_batch(
            options.model_file, 
            image_file_names,
            confidence_threshold=options.json_confidence_threshold,
            n_cores=options.n_cores,
            class_mapping_filename=options.class_mapping_filename,
            quiet=True,
            augment=options.augment,
            image_size=options.image_size)
    
        _add_frame_numbers_to_results(results)
        
        run_detector_batch.write_results_to_file(
            results, frames_json,
            relative_path_base=frame_output_folder,
            detector_file=options.model_file,
            custom_metadata={'video_frame_rate':Fs})
    
    
    ## Convert frame-level results to video-level results

    print('Converting frame-level results to video-level results')
    frame_results_to_video_results(frames_json,video_json)


    ## (Optionally) render output videos
    
    if options.render_output_video:
                
        # Render detections to images
        if (caller_provided_rendering_output_folder):
            rendering_output_dir = options.frame_rendering_folder
        else:
            rendering_output_dir = temporary_folder_info['rendering_output_folder']
            
        os.makedirs(rendering_output_dir,exist_ok=True)
        
        detected_frame_files = visualize_detector_output.visualize_detector_output(
            detector_output_path=frames_json,
            out_dir=rendering_output_dir,
            images_dir=frame_output_folder,
            confidence_threshold=options.rendering_confidence_threshold,
            preserve_path_structure=True,
            output_image_width=-1)
        detected_frame_files = [s.replace('\\','/') for s in detected_frame_files]
        
        # Choose an output folder
        output_folder_is_input_folder = False
        if options.output_video_file is not None:
            if os.path.isfile(options.output_video_file):
                raise ValueError('Rendering videos for a folder, but an existing file was specified as output')
            elif options.output_video_file == options.input_video_file:
                output_folder_is_input_folder = True
                output_video_folder = options.input_video_file
            else:
                os.makedirs(options.output_video_file,exist_ok=True)
                output_video_folder = options.output_video_file
        else:
            output_folder_is_input_folder = True
            output_video_folder = options.input_video_file
                                
        # For each video
        #
        # TODO: parallelize this loop
        #
        # i_video=0; input_video_file_abs = video_filenames[i_video]
        for i_video,input_video_file_abs in enumerate(video_filenames):
            
            video_fs = Fs[i_video]
            
            if options.rendering_fs is not None:
                rendering_fs = options.rendering_fs
            elif options.frame_sample is None:                
                rendering_fs = video_fs
            else:
                # If the original video was 30fps and we sampled every 10th frame, 
                # render at 3fps                
                rendering_fs = video_fs / options.frame_sample
            
            input_video_file_relative = os.path.relpath(input_video_file_abs,options.input_video_file)
            video_frame_output_folder = os.path.join(rendering_output_dir,input_video_file_relative)
            
            video_frame_output_folder = video_frame_output_folder.replace('\\','/')
            assert os.path.isdir(video_frame_output_folder), \
                'Could not find frame folder for video {}'.format(input_video_file_relative)
            
            # Find the corresponding rendered frame folder
            video_frame_files = [fn for fn in detected_frame_files if \
                                 fn.startswith(video_frame_output_folder)]
            assert len(video_frame_files) > 0, 'Could not find rendered frames for video {}'.format(
                input_video_file_relative)
            
            # Select the output filename for the rendered video
            if output_folder_is_input_folder:
                video_output_file = insert_before_extension(input_video_file_abs,'annotated','_')
            else:
                video_output_file = os.path.join(output_video_folder,input_video_file_relative)
            
            os.makedirs(os.path.dirname(video_output_file),exist_ok=True)
            
            # Create the output video            
            print('Rendering detections for video {} to {} at {} fps (original video {} fps)'.format(
                input_video_file_relative,video_output_file,rendering_fs,video_fs))
            frames_to_video(video_frame_files, 
                            rendering_fs, 
                            video_output_file, 
                            codec_spec=options.fourcc)
                
        # ...for each video
        
        # Possibly clean up rendered frames
        _clean_up_rendered_frames(options,rendering_output_dir,detected_frame_files)          
        
    # ...if we're rendering video
    
    
    ## (Optionally) delete the extracted frames
    _clean_up_extracted_frames(options, frame_output_folder, image_file_names)

# ...process_video_folder()


def options_to_command(options):
    """
    Convert a ProcessVideoOptions obejct to a corresponding command line.
    
    Args:
        options (ProcessVideoOptions): the options set to render as a command line
        
    Returns:
        str: the command line coresponding to [options]
        
    :meta private:
    """
    cmd = 'python process_video.py'
    cmd += ' "' + options.model_file + '"'
    cmd += ' "' + options.input_video_file + '"'
    
    if options.recursive:
        cmd += ' --recursive'
    if options.frame_folder is not None:
        cmd += ' --frame_folder' + ' "' + options.frame_folder + '"'
    if options.frame_rendering_folder is not None:
        cmd += ' --frame_rendering_folder' + ' "' + options.frame_rendering_folder + '"'    
    if options.output_json_file is not None:
        cmd += ' --output_json_file' + ' "' + options.output_json_file + '"'
    if options.output_video_file is not None:
        cmd += ' --output_video_file' + ' "' + options.output_video_file + '"'
    if options.keep_extracted_frames:
        cmd += ' --keep_extracted_frames'
    if options.reuse_results_if_available:
        cmd += ' --reuse_results_if_available'    
    if options.reuse_frames_if_available:
        cmd += ' --reuse_frames_if_available'
    if options.render_output_video:
        cmd += ' --render_output_video'
    if options.keep_rendered_frames:
        cmd += ' --keep_rendered_frames'    
    if options.rendering_confidence_threshold is not None:
        cmd += ' --rendering_confidence_threshold ' + str(options.rendering_confidence_threshold)
    if options.json_confidence_threshold is not None:
        cmd += ' --json_confidence_threshold ' + str(options.json_confidence_threshold)
    if options.n_cores is not None:
        cmd += ' --n_cores ' + str(options.n_cores)
    if options.frame_sample is not None:
        cmd += ' --frame_sample ' + str(options.frame_sample)
    if options.frames_to_extract is not None:
        cmd += ' --frames_to_extract '
        if isinstance(options.frames_to_extract,int):
            frames_to_extract = [options.frames_to_extract]
        else:
            frames_to_extract = options.frames_to_extract
        for frame_number in frames_to_extract:
            cmd += ' {}'.format(frame_number)
    if options.debug_max_frames is not None:
        cmd += ' --debug_max_frames ' + str(options.debug_max_frames)
    if options.class_mapping_filename is not None:
        cmd += ' --class_mapping_filename ' + str(options.class_mapping_filename)
    if options.fourcc is not None:
        cmd += ' --fourcc ' + options.fourcc
    if options.quality is not None:
        cmd += ' --quality ' + str(options.quality)
    if options.max_width is not None:
        cmd += ' --max_width ' + str(options.max_width)
    if options.verbose:
        cmd += ' --verbose'
    if options.force_extracted_frame_folder_deletion:
        cmd += ' --force_extracted_frame_folder_deletion'
    if options.force_rendered_frame_folder_deletion:
        cmd += ' --force_rendered_frame_folder_deletion'

    return cmd

    
#%% Interactive driver

if False:    
        
    pass

    #%% Process a folder of videos
    
    model_file = 'MDV5A'
    # input_dir = r'g:\temp\test-videos'
    input_dir = r'G:\temp\md-test-package\md-test-images\video-samples'
    output_base = r'g:\temp\video_test'
    frame_folder = os.path.join(output_base,'frames')
    rendering_folder = os.path.join(output_base,'rendered-frames')
    output_json_file = os.path.join(output_base,'video-test.json')
    output_video_folder = os.path.join(output_base,'output_videos')
    
    
    print('Processing folder {}'.format(input_dir))
    
    options = ProcessVideoOptions()    
    options.model_file = model_file
    options.input_video_file = input_dir
    options.output_video_file = output_video_folder
    options.output_json_file = output_json_file
    options.recursive = True
    options.reuse_frames_if_available = False
    options.reuse_results_if_available = False
    options.quality = 90
    options.frame_sample = 10
    options.max_width = 1280
    options.n_cores = 4
    options.verbose = True
    options.render_output_video = True    
    options.frame_folder = frame_folder
    options.frame_rendering_folder = rendering_folder    
    options.keep_extracted_frames = True
    options.keep_rendered_frames = True
    options.force_extracted_frame_folder_deletion = False
    options.force_rendered_frame_folder_deletion = False
    options.fourcc = 'mp4v'
    # options.rendering_confidence_threshold = 0.15
    
    cmd = options_to_command(options); print(cmd)
        
    # import clipboard; clipboard.copy(cmd)
    # process_video_folder(options)
        
    
    #%% Process a single video

    fn = r'g:\temp\test-videos\person_and_dog\DSCF0056.AVI'
    assert os.path.isfile(fn)
    model_file = 'MDV5A'
    input_video_file = fn
    
    output_base = r'g:\temp\video_test'
    frame_folder = os.path.join(output_base,'frames')
    rendering_folder = os.path.join(output_base,'rendered-frames')
    output_json_file = os.path.join(output_base,'video-test.json')
    output_video_file = os.path.join(output_base,'output_video.mp4')
    
    options = ProcessVideoOptions()
    options.model_file = model_file
    options.input_video_file = input_video_file
    options.render_output_video = True
    options.output_video_file = output_video_file
    options.output_json_file = output_json_file    
    options.verbose = True    
    options.quality = 75
    options.frame_sample = 10
    options.max_width = 1600    
    options.frame_folder = frame_folder
    options.frame_rendering_folder = rendering_folder    
    options.keep_extracted_frames = False
    options.keep_rendered_frames = False
    options.force_extracted_frame_folder_deletion = True
    options.force_rendered_frame_folder_deletion = True    
    options.fourcc = 'mp4v'
    # options.rendering_confidence_threshold = 0.15
    
    cmd = options_to_command(options); print(cmd)
    
    # import clipboard; clipboard.copy(cmd)    
    process_video(options)
            
    
    #%% Extract specific frames from a single video, no detection

    fn = r'g:\temp\test-videos\person_and_dog\DSCF0064.AVI'
    assert os.path.isfile(fn)
    model_file = 'no_detection'
    input_video_file = fn
    
    output_base = r'g:\temp\video_test'
    frame_folder = os.path.join(output_base,'frames')
    output_video_file = os.path.join(output_base,'output_videos.mp4')
    
    options = ProcessVideoOptions()
    options.model_file = model_file
    options.input_video_file = input_video_file    
    options.verbose = True    
    options.quality = 90
    options.frame_sample = None
    options.frames_to_extract = [0,100]
    options.max_width = None    
    options.frame_folder = frame_folder
    options.keep_extracted_frames = True
    
    cmd = options_to_command(options); print(cmd)
    
    # import clipboard; clipboard.copy(cmd)
    process_video(options)    
        
        
    #%% Extract specific frames from a folder, no detection

    fn = r'g:\temp\test-videos\person_and_dog'
    assert os.path.isdir(fn)
    model_file = 'no_detection'
    input_video_file = fn
    
    output_base = r'g:\temp\video_test'
    frame_folder = os.path.join(output_base,'frames')
    output_video_file = os.path.join(output_base,'output_videos.mp4')
    
    options = ProcessVideoOptions()
    options.model_file = model_file
    options.input_video_file = input_video_file    
    options.verbose = True    
    options.quality = 90
    options.frame_sample = None
    options.frames_to_extract = [0,100]
    options.max_width = None    
    options.frame_folder = frame_folder
    options.keep_extracted_frames = True
    
    cmd = options_to_command(options); print(cmd)
    
    # import clipboard; clipboard.copy(cmd)
    process_video(options)    

        
#%% Command-line driver

def main():

    default_options = ProcessVideoOptions()
    
    parser = argparse.ArgumentParser(description=(
        'Run MegaDetector on each frame (or every Nth frame) in a video (or folder of videos), optionally '\
        'producing a new video with detections annotated'))

    parser.add_argument('model_file', type=str,
                        help='MegaDetector model file (.pt or .pb) or model name (e.g. "MDV5A"), '\
                             'or the string "no_detection" to run just frame extraction')

    parser.add_argument('input_video_file', type=str,
                        help='video file (or folder) to process')

    parser.add_argument('--recursive', action='store_true',
                        help='recurse into [input_video_file]; only meaningful if a folder '\
                         'is specified as input')
    
    parser.add_argument('--frame_folder', type=str, default=None,
                        help='folder to use for intermediate frame storage, defaults to a folder '\
                        'in the system temporary folder')
        
    parser.add_argument('--frame_rendering_folder', type=str, default=None,
                        help='folder to use for rendered frame storage, defaults to a folder in '\
                        'the system temporary folder')
    
    parser.add_argument('--output_json_file', type=str,
                        default=None, help='.json output file, defaults to [video file].json')

    parser.add_argument('--output_video_file', type=str,
                        default=None, help='video output file (or folder), defaults to '\
                            '[video file].mp4 for files, or [video file]_annotated for folders')

    parser.add_argument('--keep_extracted_frames',
                       action='store_true', help='Disable the deletion of extracted frames')
    
    parser.add_argument('--reuse_frames_if_available',
                       action='store_true', help="Don't extract frames that are already available in the frame extraction folder")
    
    parser.add_argument('--reuse_results_if_available',
                       action='store_true', help='If the output .json files exists, and this flag is set,'\
                           'we\'ll skip running MegaDetector')
    
    parser.add_argument('--render_output_video', action='store_true',
                        help='enable video output rendering (not rendered by default)')

    parser.add_argument('--fourcc', default=default_fourcc,
                        help='fourcc code to use for video encoding (default {}), only used if render_output_video is True'.format(
                            default_fourcc))
    
    parser.add_argument('--keep_rendered_frames',
                       action='store_true', help='Disable the deletion of rendered (w/boxes) frames')

    parser.add_argument('--force_extracted_frame_folder_deletion',
                       action='store_true', help='By default, when keep_extracted_frames is False, we '\
                           'delete the frames, but leave the (probably-empty) folder in place.  This option '\
                           'forces deletion of the folder as well.  Use at your own risk; does not check '\
                           'whether other files were present in the folder.')
        
    parser.add_argument('--force_rendered_frame_folder_deletion',
                       action='store_true', help='By default, when keep_rendered_frames is False, we '\
                           'delete the frames, but leave the (probably-empty) folder in place.  This option '\
                           'forces deletion of the folder as well.  Use at your own risk; does not check '\
                           'whether other files were present in the folder.')
        
    parser.add_argument('--rendering_confidence_threshold', type=float,
                        default=None, 
                        help="don't render boxes with confidence below this threshold (defaults to choosing based on the MD version)")

    parser.add_argument('--rendering_fs', type=float,
                        default=None, 
                        help='force a specific frame rate for output videos (only relevant when using '\
                             '--render_output_video) (defaults to the original frame rate)')

    parser.add_argument('--json_confidence_threshold', type=float,
                        default=default_options.json_confidence_threshold, 
                        help="don't include boxes in the .json file with confidence "\
                            'below this threshold (default {})'.format(
                                default_options.json_confidence_threshold))

    parser.add_argument('--n_cores', type=int,
                        default=default_options.n_cores,
                        help='Number of cores to use for frame separation and detection. '\
                            'If using a GPU, this option will be respected for frame separation but '\
                            'ignored for detection.  Only relevant to frame separation when processing '\
                            'a folder.  Default {}.'.format(default_options.n_cores))

    parser.add_argument('--frame_sample', type=int,
                        default=None, help='process every Nth frame (defaults to every frame)')

    parser.add_argument('--frames_to_extract', nargs='+', type=int,
                        default=None, help='extract specific frames (one or more ints)')

    parser.add_argument('--quality', type=int,
                        default=default_options.quality, 
                        help='JPEG quality for extracted frames (defaults to {})'.format(
                            default_options.quality))

    parser.add_argument('--max_width', type=int,
                        default=default_options.max_width, 
                        help='Resize frames larger than this before writing (defaults to {})'.format(
                            default_options.max_width))

    parser.add_argument('--debug_max_frames', type=int,
                        default=-1, help='Trim to N frames for debugging (impacts model execution, '\
                            'not frame rendering)')
    
    parser.add_argument('--class_mapping_filename',
                        type=str,
                        default=None, help='Use a non-default class mapping, supplied in a .json file '\
                            'with a dictionary mapping int-strings to strings.  This will also disable '\
                            'the addition of "1" to all category IDs, so your class mapping should start '\
                            'at zero.')

    parser.add_argument('--verbose', action='store_true',
                        help='Enable additional debug output')
    
    parser.add_argument('--image_size',
                        type=int,
                        default=None,
                        help=('Force image resizing to a specific integer size on the long '\
                              'axis (not recommended to change this)'))    
    
    parser.add_argument('--augment',
                        action='store_true',
                        help='Enable image augmentation')
            
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()
        
    args = parser.parse_args()
    options = ProcessVideoOptions()
    args_to_object(args,options)

    if os.path.isdir(options.input_video_file):
        process_video_folder(options)
    else:
        process_video(options)

if __name__ == '__main__':
    main()
